unit ladybug_hw;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     z80,main_engine,sn_76496,controls_engine,gfx_engine,rom_engine,cpu_engine,
     pal_engine,sound_engine,misc_functions;

procedure Cargar_ladybug;
procedure ladybug_principal;
function iniciar_ladybug:boolean;
procedure reset_ladybug;
procedure cerrar_ladybug;
//Main CPU
function ladybug_getbyte(direccion:word):byte;
procedure ladybug_putbyte(direccion:word;valor:byte);
procedure ladybug_sonido;

const
        ladybug_rom:array[0..6] of tipo_roms=(
        (n:'l1.c4';l:$1000;p:0;crc:$d09e0adb),(n:'l2.d4';l:$1000;p:$1000;crc:$88bc4a0a),
        (n:'l3.e4';l:$1000;p:$2000;crc:$53e9efce),(n:'l4.h4';l:$1000;p:$3000;crc:$ffc424d7),
        (n:'l5.j4';l:$1000;p:$4000;crc:$ad6af809),(n:'l6.k4';l:$1000;p:$5000;crc:$cf1acca4),());
        ladybug_pal:array[0..2] of tipo_roms=(
        (n:'10-2.k1';l:$20;p:0;crc:$df091e52),(n:'10-1.f4';l:$20;p:$20;crc:$40640d8f),());
        ladybug_char:array[0..2] of tipo_roms=(
        (n:'l9.f7';l:$1000;p:0;crc:$77b1da1e),(n:'l0.h7';l:$1000;p:$1000;crc:$aa82e00b),());
        ladybug_sprites:array[0..2] of tipo_roms=(
        (n:'l8.l7';l:$1000;p:0;crc:$8b99910b),(n:'l7.m7';l:$1000;p:$1000;crc:$86a5b448),());
var
 irq_vblank:boolean;

implementation
uses principal;

procedure Cargar_ladybug;
begin
llamadas_maquina.iniciar:=iniciar_ladybug;
llamadas_maquina.bucle_general:=ladybug_principal;
llamadas_maquina.cerrar:=cerrar_ladybug;
llamadas_maquina.reset:=reset_ladybug;
end;

function iniciar_ladybug:boolean;
var
      colores:tpaleta;
      f:word;
      bit0,bit1:byte;
      memoria_temp:array[0..$1fff] of byte;
      rweights,gweights,bweights:array[0..1] of double;
const
  ps_x:array[0..15] of dword=(0, 2, 4, 6, 8, 10, 12, 14,
			8*16+0, 8*16+2, 8*16+4, 8*16+6, 8*16+8, 8*16+10, 8*16+12, 8*16+14);
  ps_y:array[0..15] of dword=(23*16, 22*16, 21*16, 20*16, 19*16, 18*16, 17*16, 16*16,
			7*16, 6*16, 5*16, 4*16, 3*16, 2*16, 1*16, 0*16);
  pc_x:array[0..7] of dword=(7, 6, 5, 4, 3, 2, 1, 0);
  pc_y:array[0..7] of dword=(0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8);
  pss_x:array[0..7] of dword=(0, 2, 4, 6, 8, 10, 12, 14);
  pss_y:array[0..7] of dword=(7*16, 6*16, 5*16, 4*16, 3*16, 2*16, 1*16, 0*16);
  resistances:array[0..1] of integer=(470,220);
begin
iniciar_ladybug:=false;
iniciar_audio(false);
//Pantallas:  principal+char y sprites
p_final[0].x:=192;
p_final[0].y:=240;
p_final[1].x:=256;
p_final[1].y:=256;
p_final[1].final_mix:=true;
p_final[2].x:=256;
p_final[2].y:=256;
p_final[2].trans:=true;
iniciar_video;
//Main CPU
init_cpu(z80_main,0,4000000/256);
main_z80_ll.getbyte:=ladybug_getbyte;
main_z80_ll.putbyte:=ladybug_putbyte;
//Audio chips
SN76496_init(0,4000000);
SN76496_init(1,4000000);
init_sound_engine(0,4000000,ladybug_sonido);
//cargar roms
if not(cargar_roms(@memoria[0],@ladybug_rom[0],'ladybug.zip',0)) then exit;
//convertir chars
if not(cargar_roms(@memoria_temp[0],@ladybug_char[0],'ladybug.zip',0)) then exit;
init_gfx(0,8,8,512);
gfx[0].trans[0]:=true;
des_gfx.pos_planos[0]:=0;
des_gfx.pos_planos[1]:=512*8*8;
des_gfx.bit_pixel:=2;
des_gfx.banks:=0;
des_gfx.long_sprites:=8*8;
convert_gfx(@gfx[0],0,@memoria_temp[0],@pc_x[0],@pc_y[0],false,true);
//convertir sprites
if not(cargar_roms(@memoria_temp[0],@ladybug_sprites[0],'ladybug.zip',0)) then exit;
init_gfx(1,16,16,128);
gfx[1].trans[0]:=true;
des_gfx.pos_planos[0]:=1;
des_gfx.pos_planos[1]:=0;
des_gfx.long_sprites:=64*8;
convert_gfx(@gfx[1],0,@memoria_temp[0],@ps_x[0],@ps_y[0],false,true);
//convetir sprites pequeos
init_gfx(2,8,8,512);
des_gfx.long_sprites:=16*8;
convert_gfx(@gfx[2],0,@memoria_temp[0],@pss_x[0],@pss_y[0],false,true);
//poner la paleta
if not(cargar_roms(@memoria_temp[0],@ladybug_pal[0],'ladybug.zip',0)) then exit;
compute_resistor_weights(0,	255, -1.0,
			2,@resistances[0],@rweights[0],470,0,
			2,@resistances[0],@gweights[0],470,0,
			2,@resistances[0],@bweights[0],470,0);
for f:=0 to $1f do begin
		// red component */
		bit0:=(not(memoria_temp[f]) shr 0) and $01;
		bit1:=(not(memoria_temp[f]) shr 5) and $01;
		colores[f].r:=combine_2_weights(@rweights[0],bit0,bit1);
		// green component */
		bit0:=(not(memoria_temp[f]) shr 2) and $01;
		bit1:=(not(memoria_temp[f]) shr 6) and $01;
		colores[f].g:=combine_2_weights(@gweights[0],bit0,bit1);
		// blue component */
		bit0:=(not(memoria_temp[f]) shr 4) and $01;
		bit1:=(not(memoria_temp[f]) shr 7) and $01;
		colores[f].b:=combine_2_weights(@bweights[0],bit0,bit1);
end;
set_pal(colores,$20);
for f:=0 to $1f do gfx[0].colores[f]:=((f shl 3) and $18) or ((f shr 2) and $07);
for f:=0 to $1f do begin
		gfx[1].colores[f]:=BITSWAP8((memoria_temp[f+$20] shr 0) and $f,7,6,5,4,0,1,2,3);
    gfx[1].colores[f+$20]:=BITSWAP8((memoria_temp[f+$20] shr 4) and $f,7,6,5,4,0,1,2,3);
end;
//final
reset_ladybug;
iniciar_ladybug:=true;
end;

procedure cerrar_ladybug;
begin
cerrar_cpus;
SN76496_close(0);
SN76496_close(1);
close_audio;
close_video;
end;

procedure reset_ladybug;
begin
 reset_z80(main_z80);
 SN76496_reset(0);
 SN76496_reset(1);
 reset_audio;
 irq_vblank:=false;
 in0:=$FF;
 in1:=$3F;
end;

procedure update_video_ladybug;inline;
var
        f,color,nchar:word;
        x,y,atrib:byte;
        flipx,flipy:boolean;
begin
fill_full_screen(1,0);
for f:=$0 to $3ff do begin
  if gfx[0].buffer[f] then begin
    x:=f div 32;
    y:=31-(f mod 32);
    atrib:=memoria[$d400+f];
    nchar:=memoria[$d000+f]+(atrib and $08) shl 5;
    color:=(atrib and $7) shl 2;
    put_gfx_trans(x*8,y*8,nchar,color,2,0);
    gfx[0].buffer[f]:=false;
  end;
end;
actualiza_trozo(0,0,256,256,2,0,0,256,256,1);
{abccdddd eeeeeeee fffghhhh iiiiiiii
 a enable?
 b size (0 = 8x8, 1 = 16x16)
 cc flip
 dddd x offset
 eeeeeeee sprite code (shift right 2 bits for 16x16 sprites)
 fff unknown
 g sprite bank
 hhhh color
 iiiiiiii y position}
for f:=$e0 downto $10 do begin
  atrib:=buffer_sprites[f*$4];
  if (atrib and $80)<>0 then begin
    color:=(buffer_sprites[$2+(f*4)] and $f) shl 2;
    y:=240-(buffer_sprites[$3+(f*4)]);
    flipy:=(atrib and $20)<>0;
    flipx:=(atrib and $10)<>0;
    if (atrib and $40)<>0	then begin // 16x16 */
      nchar:=(buffer_sprites[$1+(f*4)] shr 2)+4*(buffer_sprites[$2+(f*4)] and $10);
      x:=((f*4) shr 2)-8+(atrib and $f);
      put_gfx_sprite(nchar,color,flipx,flipy,1);
      actualiza_gfx_sprite(x,y,1,1);
    end else begin  //8x8 Parece ser que LB no usa los sprites pequeos!!!
      nchar:=buffer_sprites[$1+(f*$4)]+16*(buffer_sprites[$2+(f*4)] and $10);
      x:=((f*$4) shr 2)+(atrib and $f);
      put_gfx_sprite(nchar,color,flipx,flipy,2);
      actualiza_gfx_sprite(x,y,1,2);
    end;
  end;
end;
actualiza_trozo_final(32,8,192,240,1);
end;

procedure eventos_ladybug;
begin
if event.arcade then begin
  if arcade_input.up[0] then in0:=(in0 and $F7) else in0:=(in0 or $8);
  if arcade_input.down[0] then in0:=(in0 and $fd) else in0:=(in0 or $2);
  if arcade_input.left[0] then in0:=(in0 and $fe) else in0:=(in0 or $1);
  if arcade_input.right[0] then in0:=(in0 and $Fb) else in0:=(in0 or $4);
  if arcade_input.but0[0] then in0:=(in0 and $ef) else in0:=(in0 or $10);
  if arcade_input.coin[0] then main_z80.pedir_nmi:=ASSERT_LINE else clear_nmi_z80(main_z80);
  if arcade_input.coin[1] then main_z80.pedir_irq:=HOLD_LINE;
  if arcade_input.start[0] then in0:=(in0 and $df) else in0:=(in0 or $20);
  if arcade_input.start[1] then in0:=(in0 and $bf) else in0:=(in0 or $40);
end;
end;

procedure ladybug_principal;
var
  frame:single;
  f:byte;
begin
init_controls(false,false,false,true);
frame:=main_z80.tframes;
while EmuStatus=EsRuning do begin
  for f:=0 to $ff do begin
    run_z80(main_z80,main_z80_ll,trunc(frame));
    frame:=frame+main_z80.tframes-main_z80.contador;
    case f of
      224:begin
            in1:=(in1 and $3f) or $80;
            update_video_ladybug;
            copymemory(@buffer_sprites[0],@memoria[$7000],$400);
          end;
      255:in1:=(in1 and $3f) or $40;
    end;
  end;
  eventos_ladybug;
  video_sync;
end;
end;

function ladybug_getbyte(direccion:word):byte;
begin
case direccion of
        $9000:ladybug_getbyte:=in0;
        $9001:ladybug_getbyte:=in1;
        $9002:ladybug_getbyte:=$df;
        $9003:ladybug_getbyte:=$ff;
        $e000:ladybug_getbyte:=$ff;
        else ladybug_getbyte:=memoria[direccion];
end;
end;

procedure ladybug_putbyte(direccion:word;valor:byte);
begin
if direccion<$6000 then exit;
memoria[direccion]:=valor;
case direccion of
        $b000..$bfff:SN76496_Write(0,valor);
        $c000..$cfff:SN76496_Write(1,valor);
        $d000..$dfff:gfx[0].buffer[direccion and $3ff]:=true;
end;
end;

procedure ladybug_sonido;
begin
  SN76496_Update(0);
  SN76496_Update(1);
end;

end.
