unit deco_common;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}gfx_engine,ym_2203,ym_2151,oki6295,hu6280,
     cpu_engine,sound_engine,main_engine;

var
 sprite_ram:array[0..$7ff] of byte;
 deco16_global_x_size,deco16_sprite_color_add,deco16_sprite_mask:word;
 deco16_sound_latch:byte;
 snd_ram:array[0..$1fff] of byte;

//Sprites
procedure deco16_sprites;
procedure deco16_sprites_pri(pri:byte);
//Sound
procedure deco16_snd_simple_init(cpu_clock,sound_clock:dword;sound_bank:porthandler);
procedure deco16_snd_simple_reset;
procedure deco16_snd_simple_close;
function deco16_simple_snd_getbyte(direccion:dword):Byte;
procedure deco16_simple_snd_putbyte(direccion:dword;valor:byte);
procedure deco16_simple_sound;
procedure deco16_snd_double_init(cpu_clock,sound_clock:dword;sound_bank:porthandler);
procedure deco16_snd_double_reset;
procedure deco16_snd_double_close;
function deco16_double_snd_getbyte(direccion:dword):Byte;
procedure deco16_double_snd_putbyte(direccion:dword;valor:byte);
procedure deco16_double_sound;
procedure deco16_snd_irq(numchip:integer;irqstate:integer);

implementation
uses principal;

//sprites
procedure deco16_sprites;inline;
var
  f,color:byte;
  y,x,nchar:word;
  fx,fy:boolean;
  multi,inc,mult:integer;
begin
for f:=0 to $ff do begin
			nchar:=sprite_ram[(f*8)+3]+(sprite_ram[(f*8)+2] shl 8);
			y:=sprite_ram[(f*8)+1]+(sprite_ram[(f*8)+0] shl 8);
      if (((y and $1000)<>0) and ((main_vars.frames_sec and 1)<>0)) then continue;
      x:=sprite_ram[(f*8)+5]+(sprite_ram[(f*8)+4] shl 8);
      color:=(x shr 9) and $1f;
      fx:=(y and $2000)<>0;
      fy:=(y and $4000)<>0;
      multi:=(1 shl ((y and $0600) shr 9))-1;
      x:=(304-x) and $1ff;
      y:=(240-y) and $1ff;
			nchar:=(nchar and not(multi)) and deco16_sprite_mask;
      if fy then inc:=-1
        else begin
						nchar:=nchar+multi;
						inc:=1;
        end;
      mult:=-16;
      while (multi>=0) do begin
        if nchar<>0 then begin
          put_gfx_sprite(nchar-multi*inc,(color shl 4)+deco16_sprite_color_add,fx,fy,2);
          actualiza_gfx_sprite(x,y+mult*multi,3,2);
        end;
        multi:=multi-1;
      end;
end;
end;

procedure deco16_sprites_pri(pri:byte);
var
  f,color:byte;
  y,x,nchar:word;
  fx,fy:boolean;
  multi,inc,mult:integer;
begin
for f:=0 to $ff do begin
      x:=buffer_sprites[(f*8)+5]+(buffer_sprites[(f*8)+4] shl 8);
      if pri<>((x shr 8) and $c0) then continue;
			nchar:=buffer_sprites[(f*8)+3]+(buffer_sprites[(f*8)+2] shl 8);
			y:=buffer_sprites[(f*8)+1]+(buffer_sprites[(f*8)+0] shl 8);
      if (((y and $1000)<>0) and ((main_vars.frames_sec and 1)<>0)) then continue;
      color:=(x shr 9) and $1f;
      fx:=(y and $2000)<>0;
      fy:=(y and $4000)<>0;
      multi:=(1 shl ((y and $0600) shr 9))-1;	// 1x, 2x, 4x, 8x height
      x:=(deco16_global_x_size-x) and $1ff;
      y:=(240-y) and $1ff;
			nchar:=(nchar and not(multi)) and deco16_sprite_mask;
      if fy then inc:=-1
        else begin
						nchar:=nchar+multi;
						inc:=1;
        end;
      mult:=-16;
      while (multi>=0) do begin
        if nchar<>0 then begin
          put_gfx_sprite(nchar-multi*inc,(color shl 4)+deco16_sprite_color_add,fx,fy,3);
          actualiza_gfx_sprite(x,y+mult*multi,5,3);
        end;
        multi:=multi-1;
      end;
end;
end;

//Sound
procedure deco16_snd_double_init(cpu_clock,sound_clock:dword;sound_bank:porthandler);
begin
init_cpu(h6280_main,1,cpu_clock/$100);
main_h6280_ll.getbyte:=deco16_double_snd_getbyte;
main_h6280_ll.putbyte:=deco16_double_snd_putbyte;
ym2203_Init(0,1,sound_clock div 8,cpu_clock,1,nil,nil,nil,nil,nil);
YM2151_Init(0,1,sound_clock div 9,cpu_clock,sound_bank,deco16_snd_irq);
init_oki6295(0,sound_clock div 32,OKIM6295_PIN7_HIGH,1,cpu_clock,2);
init_oki6295(1,sound_clock div 16,OKIM6295_PIN7_HIGH,1,cpu_clock,2);
init_sound_engine(1,cpu_clock,deco16_double_sound);
end;

procedure deco16_snd_double_reset;
begin
  reset_h6280(main_h6280,main_h6280_ll);
  ym2203_reset(0);
  YM2151_reset(0);
  reset_oki6295(0);
  reset_oki6295(1);
  deco16_sound_latch:=0;
end;

procedure deco16_snd_double_close;
begin
ym2203_close(0);
YM2151_close(0);
close_oki6295(0);
close_oki6295(1);
end;

function deco16_double_snd_getbyte(direccion:dword):Byte;
begin
case direccion of
  0..$ffff:deco16_double_snd_getbyte:=mem_snd[direccion];
  $100000:deco16_double_snd_getbyte:=ym2203_Read_Status(0);
  $100001:deco16_double_snd_getbyte:=ym2203_Read_Reg(0);
  $110001:deco16_double_snd_getbyte:=YM2151_status_port_read(0);
  $120000..$120001:deco16_double_snd_getbyte:=okim6295_r(0);
  $130000..$130001:deco16_double_snd_getbyte:=okim6295_r(1);
  $140000..$140001:deco16_double_snd_getbyte:=deco16_sound_latch;
  $1f0000..$1f1fff:deco16_double_snd_getbyte:=snd_ram[direccion and $1fff];
end;
end;

procedure deco16_double_snd_putbyte(direccion:dword;valor:byte);
begin
if direccion<$10000 then exit;
case direccion of
  $100000:ym2203_Control(0,valor);
  $100001:ym2203_Write_Reg(0,valor);
  $110000:YM2151_register_port_write(0,valor);
  $110001:YM2151_data_port_write(0,valor);
  $120000..$120001:okim6295_w(0,valor);
  $130000..$130001:okim6295_w(1,valor);
  $1f0000..$1f1fff:snd_ram[direccion and $1fff]:=valor;
  $1fec00..$1fec01:h6280_timer_w(main_h6280,direccion and $1,valor);
  $1ff400..$1ff403:h6280_irq_status_w(main_h6280,direccion and $3,valor);
end;
end;

procedure deco16_double_sound;
begin
  ym2151_Update(0);
  ym2203_Update(0);
end;

procedure deco16_snd_simple_init(cpu_clock,sound_clock:dword;sound_bank:porthandler);
begin
  init_cpu(h6280_main,1,cpu_clock/$100);
  main_h6280_ll.getbyte:=deco16_simple_snd_getbyte;
  main_h6280_ll.putbyte:=deco16_simple_snd_putbyte;
  YM2151_Init(0,1,sound_clock div 9,cpu_clock,sound_bank,deco16_snd_irq);
  init_oki6295(0,sound_clock div 32,OKIM6295_PIN7_HIGH,1,cpu_clock,2);
  init_sound_engine(1,cpu_clock,deco16_simple_sound);
end;

procedure deco16_snd_simple_reset;
begin
  reset_h6280(main_h6280,main_h6280_ll);
  YM2151_reset(0);
  reset_oki6295(0);
  deco16_sound_latch:=0;
end;

procedure deco16_snd_simple_close;
begin
  YM2151_close(0);
  close_oki6295(0);
end;

function deco16_simple_snd_getbyte(direccion:dword):Byte;
begin
case direccion of
  0..$ffff:deco16_simple_snd_getbyte:=mem_snd[direccion];
  $110001:deco16_simple_snd_getbyte:=YM2151_status_port_read(0);
  $120000..$120001:deco16_simple_snd_getbyte:=okim6295_r(0);
  $140000..$140001:deco16_simple_snd_getbyte:=deco16_sound_latch;
  $1f0000..$1f1fff:deco16_simple_snd_getbyte:=snd_ram[direccion and $1fff];
end;
end;

procedure deco16_simple_snd_putbyte(direccion:dword;valor:byte);
begin
if direccion<$10000 then exit;
case direccion of
  $110000:YM2151_register_port_write(0,valor);
  $110001:YM2151_data_port_write(0,valor);
  $120000..$120001:okim6295_w(0,valor);
  $1f0000..$1f1fff:snd_ram[direccion and $1fff]:=valor;
  $1fec00..$1fec01:h6280_timer_w(main_h6280,direccion and $1,valor);
  $1ff400..$1ff403:h6280_irq_status_w(main_h6280,direccion and $3,valor);
end;
end;

procedure deco16_snd_irq(numchip:integer;irqstate:integer);
begin
  if irqstate<>0 then set_irq_line(main_h6280,1,ASSERT_LINE)
    else set_irq_line(main_h6280,1,CLEAR_LINE);
end;

procedure deco16_simple_sound;
begin
  ym2151_Update(0);
end;

end.
