unit snowbros_hw;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     z80,m68000,main_engine,controls_engine,gfx_engine,ym_3812,
     rom_engine,cpu_engine,pal_engine,kaneco_pandora,sound_engine;

procedure Cargar_snowbros;
procedure snowbros_principal;
function iniciar_snowbros:boolean;
procedure reset_snowbros;
procedure cerrar_snowbros;
//Main CPU
function snowbros_getword(direccion:dword):word;
procedure snowbros_putword(direccion:dword;valor:word);
//Sound CPU
function snowbros_snd_getbyte(direccion:word): Byte;
procedure snowbros_snd_putbyte(direccion:word;valor:byte);
procedure snowbros_sound_act;
function snowbros_snd_inbyte(puerto:word):byte;
procedure snowbros_snd_outbyte(valor:byte;puerto:word);
procedure snd_irq(numchip:integer;irqstate:integer);

const
        snowbros_rom:array[0..2] of tipo_roms=(
        (n:'sn6.bin';l:$20000;p:0;crc:$4899ddcf),(n:'sn5.bin';l:$20000;p:$1;crc:$ad310d3f),());
        snowbros_char:tipo_roms=(n:'sbros-1.41';l:$80000;p:0;crc:$16f06b3a);
        snowbros_sound:tipo_roms=(n:'sbros-4.29';l:$8000;p:0;crc:$e6eab4e4);

var
 rom:array[0..$3ffff] of byte;
 ram:array[0..$3fff] of byte;
 sound_latch:byte;

implementation
uses principal;

procedure Cargar_snowbros;
begin
llamadas_maquina.iniciar:=iniciar_snowbros;
llamadas_maquina.bucle_general:=snowbros_principal;
llamadas_maquina.cerrar:=cerrar_snowbros;
llamadas_maquina.reset:=reset_snowbros;
llamadas_maquina.fps_max:=57.5;
end;

function iniciar_snowbros:boolean; 
const
  pc_x:array[0..15] of dword=(0, 4, 8, 12, 16, 20, 24, 28,
		8*32+0, 8*32+4, 8*32+8, 8*32+12, 8*32+16, 8*32+20, 8*32+24, 8*32+28);
  pc_y:array[0..15] of dword=(0*32, 1*32, 2*32, 3*32, 4*32, 5*32, 6*32, 7*32,
		16*32, 17*32, 18*32, 19*32, 20*32, 21*32, 22*32, 23*32);
var
  memoria_temp:array[0..$7ffff] of byte;
begin
iniciar_snowbros:=false;
iniciar_audio(false);
//Pantallas:  principal+char y sprites
p_final[0].x:=256;
p_final[0].y:=224;
p_final[1].x:=512;
p_final[1].y:=512;
p_final[1].trans:=true;
p_final[1].final_mix:=true;
iniciar_video;
//cargar roms
if not(cargar_roms16b(@rom[0],@snowbros_rom[0],'snowbros.zip',0)) then exit;
//cargar sonido
if not(cargar_roms(@mem_snd[0],@snowbros_sound,'snowbros.zip',1)) then exit;
//Main CPU
init_cpu(m68000_main,0,8000000/3);
main_68000_ll.getword:=snowbros_getword;
main_68000_ll.putword:=snowbros_putword;
//Sound CPU
init_cpu(z80_snd,1,6000000/3);
snd_z80_ll.getbyte:=snowbros_snd_getbyte;
snd_z80_ll.putbyte:=snowbros_snd_putbyte;
snd_z80_ll.outbyte:=snowbros_snd_outbyte;
snd_z80_ll.inbyte:=snowbros_snd_inbyte;
//pandora
pandora.mask_nchar:=$fff;
pandora.color_offset:=0;
pandora.clear_screen:=true;
//Sound Chips
ym3812_init(0,1,3000000,6000000,snd_irq,1);
init_sound_engine(1,6000000,snowbros_sound_act);
//convertir chars
if not(cargar_roms(@memoria_temp[0],@snowbros_char,'snowbros.zip',1)) then exit;
init_gfx(0,16,16,$1000);
gfx[0].trans[0]:=true;
des_gfx.pos_planos[0]:=0;
des_gfx.pos_planos[1]:=1;
des_gfx.pos_planos[2]:=2;
des_gfx.pos_planos[3]:=3;
des_gfx.bit_pixel:=4;
des_gfx.banks:=0;
des_gfx.long_sprites:=32*32;
convert_gfx(@gfx[0],0,@memoria_temp[0],@pc_x[0],@pc_y[0],false,false);
//final
reset_snowbros;
iniciar_snowbros:=true;
end;

procedure cerrar_snowbros;
begin
cerrar_cpus;
ym3812_close(0);
close_audio;
close_video;
end;

procedure reset_snowbros;
begin
 reset_68000(main_68000,main_68000_ll);
 reset_z80(snd_z80);
 pandora_reset;
 ym3812_reset(0);
 reset_audio;
 in0:=$FF;
 in1:=$7F;
 in2:=$7F;
 sound_latch:=0;
end;

procedure update_video_snowbros;inline;
begin
pandora_update_video(1,0);
actualiza_trozo_final(0,16,256,224,1);
end;

procedure eventos_snowbros;
begin
if event.arcade then begin
  if arcade_input.up[0] then in1:=(in1 and $fe) else in1:=(in1 or $1);
  if arcade_input.down[0] then in1:=(in1 and $Fd) else in1:=(in1 or $2);
  if arcade_input.left[0] then in1:=(in1 and $fb) else in1:=(in1 or $4);
  if arcade_input.right[0] then in1:=(in1 and $F7) else in1:=(in1 or $8);
  if arcade_input.but1[0] then in1:=(in1 and $ef) else in1:=(in1 or $10);
  if arcade_input.but0[0] then in1:=(in1 and $df) else in1:=(in1 or $20);
  if arcade_input.but2[0] then in1:=(in1 and $bf) else in1:=(in1 or $40);
  if arcade_input.coin[0] then in0:=(in0 and $fb) else in0:=(in0 or $4);
  if arcade_input.coin[1] then in0:=(in0 and $f7) else in0:=(in0 or $8);
  if arcade_input.start[0] then in0:=(in0 and $fe) else in0:=(in0 or $1);
  if arcade_input.start[1] then in0:=(in0 and $fd) else in0:=(in0 or $2);
  if arcade_input.up[1] then in2:=(in2 and $fe) else in2:=(in2 or $1);
  if arcade_input.down[1] then in2:=(in2 and $Fd) else in2:=(in2 or $2);
  if arcade_input.left[1] then in2:=(in2 and $fb) else in2:=(in2 or $4);
  if arcade_input.right[1] then in2:=(in2 and $F7) else in2:=(in2 or $8);
  if arcade_input.but1[1] then in2:=(in2 and $ef) else in2:=(in2 or $10);
  if arcade_input.but0[1] then in2:=(in2 and $df) else in2:=(in2 or $20);
  if arcade_input.but2[1] then in2:=(in2 and $bf) else in2:=(in2 or $40);
end;
end;

procedure snowbros_principal;
var
  frame_m,frame_s:single;
  f:byte;
begin
init_controls(false,false,false,true);
frame_m:=main_68000.tframes;
frame_s:=snd_z80.tframes;
while EmuStatus=EsRuning do begin
 for f:=0 to 2 do begin
  //Main CPU
  run_68000(main_68000,main_68000_ll,trunc(frame_m));
  frame_m:=frame_m+main_68000.tframes-main_68000.contador;
  main_68000.irq[4-f]:=ASSERT_LINE;
  //Sound CPU
  run_z80(snd_z80,snd_z80_ll,trunc(frame_s));
  frame_s:=frame_s+snd_z80.tframes-snd_z80.contador;
 end;
 update_video_snowbros;
 eventos_snowbros;
 video_sync;
end;
end;

procedure cambiar_color(tmp_color,numero:word);inline;
var
  color:tcolor;
begin
  color.b:=pal5bit(tmp_color shr 10);
  color.g:=pal5bit(tmp_color shr 5);
  color.r:=pal5bit(tmp_color);
  set_pal_color(color,addr(paleta[numero]));
end;

function snowbros_getword(direccion:dword):word;
begin
direccion:=direccion and $fffffe;
case direccion of
    0..$3ffff:snowbros_getword:=(rom[direccion] shl 8)+rom[direccion+1];
    $100000..$103fff:snowbros_getword:=(ram[direccion and $3fff] shl 8)+ram[(direccion+1) and $3fff];
    $300000:snowbros_getword:=sound_latch;
    $500000:snowbros_getword:=(in1 shl 8)+$fe;
    $500002:snowbros_getword:=(in2 shl 8)+$ff;
    $500004:snowbros_getword:=(in0 shl 8)+$00;
    $600000..$6001ff:snowbros_getword:=buffer_paleta[(direccion and $1ff)+1] or (buffer_paleta[direccion and $1ff] shl 8);
    $700000..$701fff:snowbros_getword:=pandora.sprite_ram[(direccion and $1fff) shr 1]
end;
end;

procedure snowbros_putword(direccion:dword;valor:word);
begin
direccion:=direccion and $fffffe;
if direccion<$40000 then exit;
case direccion of
    $100000..$103fff:begin
                        ram[direccion and $3fff]:=valor shr 8;
                        ram[(direccion+1) and $3fff]:=valor and $ff;
                     end;
    $200000,$400000,$702000..$7022ff:exit;
    $300000:begin
            sound_latch:=valor and $ff;
            snd_z80.pedir_nmi:=PULSE_LINE;
          end;
    $600000..$6001ff:begin
                      buffer_paleta[direccion and $1ff]:=valor shr 8;
                      buffer_paleta[(direccion+1) and $1ff]:=valor and $ff;
                      cambiar_color(valor,((direccion and $1ff) shr 1));
                   end;
    $700000..$701fff:pandora.sprite_ram[(direccion and $1fff) shr 1]:=valor and $ff;
    $800000:main_68000.irq[4]:=CLEAR_LINE;
    $900000:main_68000.irq[3]:=CLEAR_LINE;
    $a00000:main_68000.irq[2]:=CLEAR_LINE;
  end;
end;

function snowbros_snd_getbyte(direccion:word):byte;
begin
snowbros_snd_getbyte:=mem_snd[direccion];
end;

procedure snowbros_snd_putbyte(direccion:word;valor:byte);
begin
if direccion>$7fff then mem_snd[direccion]:=valor;
end;

function snowbros_snd_inbyte(puerto:word):byte;
begin
  case (puerto and $ff) of
  $2:snowbros_snd_inbyte:=ym3812_status_port(0);
  $4:snowbros_snd_inbyte:=sound_latch;
  end;
end;

procedure snowbros_snd_outbyte(valor:byte;puerto:word);
begin
case (puerto and $ff) of
  $2:ym3812_control_port(0,valor);
  $3:ym3812_write_port(0,valor);
  $4:sound_latch:=valor;
end;
end;

procedure snowbros_sound_act;
begin
  ym3812_Update(0);
end;

procedure snd_irq(numchip:integer;irqstate:integer);
begin
  if (irqstate=1) then snd_z80.pedir_irq:=ASSERT_LINE
    else snd_z80.pedir_irq:=CLEAR_LINE;
end;

end.
