unit coleco;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses z80,sdl,{$IFDEF WINDOWS}windows,{$ENDIF}
     lenguaje,main_engine,controls_engine,tms99xx,sn_76496,sysutils,dialogs,
     rom_engine,cpu_engine,misc_functions,sound_engine,file_engine,pal_engine;

procedure Cargar_coleco;
procedure coleco_principal;
function coleco_getbyte(direccion:word): Byte;
procedure coleco_putbyte(direccion:word;valor:byte);
function coleco_inbyte(puerto:word):byte;
procedure coleco_outbyte(valor:byte;puerto:word);
function iniciar_coleco:boolean;
procedure reset_coleco;
procedure cerrar_coleco;
procedure coleco_sound;
function abrir_coleco:boolean;
procedure coleco_interrupt(int:boolean);
procedure coleco_grabar_snapshot;

const
        coleco_bios:tipo_roms=(n:'coleco.rom';l:$2000;p:0;crc:$3aa93ef3);
        keycodes:array[0..15] of byte=($0A,$0D,$07,$0C,$02,$03,$0E,$05,$01,$0B,$06,$09,$0F,$0F,$0F,$0F);
var
  njoymode:boolean;
  nJoyState:array[0..1] of Integer;

implementation
uses principal;

procedure Cargar_coleco;
begin
form1.Panel2.Visible:=true;
form1.BitBtn9.visible:=false;
form1.BitBtn10.Enabled:=true;
form1.BitBtn10.Glyph:=nil;
form1.imagelist2.GetBitmap(4,form1.BitBtn10.Glyph);
form1.BitBtn10.visible:=true;
form1.BitBtn10.OnClick:=form1.fLoadCartucho;
form1.BitBtn11.visible:=true;
form1.BitBtn11.Enabled:=true;
form1.BitBtn12.visible:=false;
form1.BitBtn14.visible:=false;
llamadas_maquina.iniciar:=iniciar_coleco;
llamadas_maquina.bucle_general:=coleco_principal;
llamadas_maquina.cerrar:=cerrar_coleco;
llamadas_maquina.reset:=reset_coleco;
llamadas_maquina.cartuchos:=abrir_coleco;
llamadas_maquina.grabar_snapshot:=coleco_grabar_snapshot;
llamadas_maquina.fps_max:=10738635/2/342/262;
end;

function iniciar_coleco:boolean;
begin
iniciar_coleco:=false;
iniciar_audio(false);
//Pantallas:  principal+char y sprites
p_final[0].x:=256+BORDER*2;
p_final[0].y:=192+BORDER*2;
iniciar_video;
//Main CPU
init_cpu(z80_main,0,3579545);
main_z80_ll.getbyte:=coleco_getbyte;
main_z80_ll.putbyte:=coleco_putbyte;
main_z80_ll.outbyte:=coleco_outbyte;
main_z80_ll.inbyte:=coleco_inbyte;
//cargar roms
if not(cargar_roms(@memoria[0],@coleco_bios,'coleco.zip',1)) then exit;
//TMS
TMS99XX_Init;
tms.IRQ_Handler:=coleco_interrupt;
//Chip Sonido
SN76496_init(0,3579545);
init_sound_engine(0,3579545,coleco_sound);
//final
reset_coleco;
iniciar_coleco:=true;
end;

procedure cerrar_coleco;
begin
SN76496_close(0);
TMS99XX_close;
cerrar_cpus;
close_audio;
close_video;
end;

procedure reset_coleco;
begin
 SN76496_reset(0);
 TMS99XX_reset;
 reset_z80(main_z80);
 reset_audio;
 fillchar(memoria[$2000],$E000,0);
 njoymode:=false;
 nJoyState[0]:=0;
 nJoyState[1]:=$FFFF;
end;

procedure coleco_principal;
var
  frame:single;
begin
init_controls(false,true,true,false);
frame:=main_z80.tframes;
while EmuStatus=EsRuning do begin
  run_z80(main_z80,main_z80_ll,trunc(frame));
  frame:=frame+main_z80.tframes-main_z80.contador;
  TMS99XX_Interrupt;
  TMS99XX_refresh;
  video_sync;
end;
end;

function coleco_getbyte(direccion:word):Byte;
begin
case direccion of
  0..$1fff,$8000..$ffff:coleco_getbyte:=memoria[direccion];
  $6000..$7fff:coleco_getbyte:=memoria[$6000+(direccion and $3ff)];
end;
end;

procedure coleco_putbyte(direccion:word;valor:byte);
begin
//Solo tiene $400 bytes de memoria RAM, hace mirror desde $6000 hasta la $7fff
if ((direccion>$5fff) and (direccion<$8000)) then memoria[$6000+(direccion and $3ff)]:=valor;
end;

function coleco_inbyte(puerto:word):byte;
var
  nPAux: Integer;
  nResult: Byte;
begin
  nResult:=$FF;
  case (puerto and $E0) of
    $E0: begin
      nJoyState[0]:=$FFFF;
      if event.arcade then begin
        if arcade_input.left[0] then nJoyState[0]:=nJoyState[0] and $F7FF;
        if arcade_input.right[0] then nJoyState[0]:=nJoyState[0] and $FDFF;
        if arcade_input.up[0] then nJoyState[0]:=nJoyState[0] and $FEFF;
        if arcade_input.down[0] then nJoyState[0]:=nJoyState[0] and $FBFF;
        if arcade_input.but1[0] then nJoyState[0]:=nJoyState[0] and $FFBF;
        if arcade_input.but0[0] then nJoyState[0]:=nJoyState[0] and $BFFF;
      end;
      if event.keyboard then begin
        if keyboard[SDLK_SPACE] then nJoyState[0]:=nJoyState[0] and $FFBF;
        if keyboard[SDLK_RETURN] then nJoyState[0]:=nJoyState[0] and $FFBF;
        if keyboard[SDLK_0] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $0;
        if keyboard[SDLK_1] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $1;
        if keyboard[SDLK_2] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $2;
        if keyboard[SDLK_3] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $3;
        if keyboard[SDLK_4] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $4;
        if keyboard[SDLK_5] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $5;
        if keyboard[SDLK_6] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $6;
        if keyboard[SDLK_7] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $7;
        if keyboard[SDLK_8] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $8;
        if keyboard[SDLK_9] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $9;
        if keyboard[SDLK_Q] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $a;
        if keyboard[SDLK_W] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $b;
      end;
      nPAux:=(puerto shr 1) and $01;
      //Para emular los dos joysticks con las mismas teclas de funcion
      //Pero no las direcciones
      nJoyState[1]:=$fff0 or nJoyState[0] and $f;
      if nJoyMode then nPAux:=nJoyState[nPAux] shr 8
        else nPAux:=(nJoyState[nPAux] and $F0) or keycodes[nJoyState[nPAux] and $0F];
      nResult:=(nPAux or $B0) and $7F;
    end;
    $A0:if (puerto and $01)<>0 then nResult:=TMS99XX_register_r
          else nResult:= TMS99XX_vram_r;
  end;
  coleco_inbyte:=nResult;
end;

procedure coleco_outbyte(valor:byte;puerto:word);
begin
  case (puerto and $E0) of
    $80:nJoyMode:=false;
    $C0:nJoyMode:=true;
    $E0:SN76496_Write(0,valor);
    $A0:if (puerto and $01)<>0 then TMS99XX_register_w(valor)
          else TMS99XX_vram_w(valor);
  end;
end;

procedure coleco_sound;
begin
  SN76496_Update(0);
end;

function coleco_cargar_snapshot(data:pbyte;long:dword):boolean;
var
  f:byte;
  cadena:string;
  longitud,comprimido,descomprimido:integer;
  ptemp:pointer;
begin
coleco_cargar_snapshot:=false;
longitud:=0;
for f:=0 to 3 do begin
  cadena:=cadena+chr(data^);
  inc(data);
end;
//Todos las cabeceras tienen 10bytes
if cadena<>'CLSN' then exit;
reset_coleco;
if data^<>1 then exit; //Version
inc(data,6);inc(longitud,10);
while longitud<>long do begin
  if longitud>long then exit;
  cadena:='';
  for f:=0 to 3 do begin
        cadena:=cadena+chr(data^);
        inc(data);inc(longitud);
  end;
  if cadena='CRAM' then begin
    comprimido:=sacar_word(data);
    inc(data,6);inc(longitud,6);
    getmem(ptemp,$e000);
    decompress_zlib(data,comprimido,ptemp,descomprimido);
    copymemory(addr(memoria[$2000]),ptemp,$e000);
    freemem(ptemp);
    inc(data,comprimido);inc(longitud,comprimido);
  end;
  if cadena='Z80R' then begin
    comprimido:=sacar_word(data);
    inc(data,6);inc(longitud,6);
    copymemory(main_z80,data,comprimido);
    inc(data,comprimido);inc(longitud,comprimido);
  end;
  if cadena='TMSR' then begin
    comprimido:=sacar_word(data);
    inc(data,2);inc(longitud,2);
    comprimido:=data^*65536+comprimido;
    inc(data,4);inc(longitud,4);
    getmem(ptemp,sizeof(TTMS99XX));
    decompress_zlib(data,comprimido,ptemp,descomprimido);
    copymemory(TMS,ptemp,descomprimido);
    freemem(ptemp);
    inc(data,comprimido);inc(longitud,comprimido);
    if tms.nBGColor=0 then paleta[0]:=0
      else paleta[0]:=paleta[tms.nBGColor];
  end;
end;
coleco_cargar_snapshot:=true;
end;

function abrir_cartucho(datos:pbyte;longitud:integer):boolean;
var
  ptemp:pbyte;
begin
abrir_cartucho:=false;
ptemp:=datos;
inc(ptemp,1);
if not(((datos^=$55) and (ptemp^=$aa)) or ((datos^=$aa) and (ptemp^=$55)) or ((datos^=$66) and (ptemp^=$99))) then exit;
reset_coleco;
copymemory(addr(memoria[$8000]),datos,longitud);
abrir_cartucho:=true;
end;

function abrir_coleco:boolean;
var
  extension,nombre_file,RomFile:string;
  datos:pbyte;
  longitud,crc:integer;
begin
  if not(OpenRom(StColecovision,Romfile)) then begin
    abrir_coleco:=true;
    exit;
  end;
  abrir_coleco:=false;
  extension:=extension_fichero(RomFile);
  if extension='ZIP' then begin
    if not(search_file_from_zip(RomFile,'*.col',nombre_file,longitud,crc,false)) then
      if not(search_file_from_zip(RomFile,'*.rom',nombre_file,longitud,crc,false)) then
        if not(search_file_from_zip(RomFile,'*.bin',nombre_file,longitud,crc,false)) then
          if not(search_file_from_zip(RomFile,'*.csn',nombre_file,longitud,crc,true)) then exit;
    getmem(datos,longitud);
    if not(load_file_from_zip(RomFile,nombre_file,datos,longitud,crc,true)) then begin
      freemem(datos);
      exit;
    end;
  end else begin
    if ((extension<>'COL') and (extension<>'ROM') and (extension<>'BIN') and (extension<>'CSN')) then exit;
    if not(read_file_size(RomFile,longitud)) then exit;
    getmem(datos,longitud);
    if not(read_file(RomFile,datos,longitud)) then begin
      freemem(datos);
      exit;
    end;
    nombre_file:=extractfilename(RomFile);
  end;
directory.ColecoVision:=ExtractFilePath(romfile);
extension:=extension_fichero(nombre_file);
if extension='CSN' then begin
  if not(coleco_cargar_snapshot(datos,longitud)) then begin
    freemem(datos);
    exit;
  end;
end;
if ((extension='COL') or (extension='ROM') or (extension='BIN')) then begin
  if not(abrir_cartucho(datos,longitud)) then begin
    freemem(datos);
    exit;
  end;
end;
freemem(datos);
change_caption(llamadas_maquina.caption+' - '+nombre_file);
//Restauro la llamada de interrupcion del TMS
tms.IRQ_Handler:=coleco_interrupt;
abrir_coleco:=true;
end;

procedure coleco_grabar_snapshot;
var
  cantidad,long_final,comprimido:integer;
  buffer:array[0..9] of byte;  //Cabecera de los bloques siempre 10bytes
  nombre:string;
  puntero,datos_final,data:pbyte;
  ptemp:pointer;
begin
form1.savedialog1.InitialDir:=Directory.coleco_snap;
form1.saveDialog1.Filter := 'CSN Format (*.csn)|*.csn';
if form1.savedialog1.execute then begin
        nombre:=changefileext(form1.savedialog1.FileName,'.csn');
        if FileExists(nombre) then begin                                         //Respuesta 'NO' es 7
            if MessageDlg(leng[main_vars.idioma].mensajes[3], mtWarning, [mbYes]+[mbNo],0)=7 then exit;
        end;
end else exit;
long_final:=0;
getmem(datos_final,139400);
data:=datos_final;
//Cabecera
buffer[0]:=ord('C'); //Nombre Bloque
buffer[1]:=ord('L');
buffer[2]:=ord('S');
buffer[3]:=ord('N');
buffer[4]:=1; //version
buffer[5]:=0;buffer[6]:=0;buffer[7]:=0;buffer[8]:=0;buffer[9]:=0; //reservado
copymemory(data,addr(buffer[0]),10);
inc(data,10);inc(long_final,10);
//Coleco RAM longitud=$e000
buffer[0]:=ord('C');
buffer[1]:=ord('R');
buffer[2]:=ord('A');
buffer[3]:=ord('M');
buffer[6]:=0;buffer[7]:=0;buffer[8]:=0;buffer[9]:=0;  //reservado
getmem(puntero,$e000);
ptemp:=addr(memoria[$2000]);
compress_zlib(ptemp,$e000,pointer(puntero),comprimido);
buffer[4]:=comprimido mod 256; //longitud
buffer[5]:=comprimido div 256; //longitud
copymemory(data,addr(buffer[0]),10);
inc(data,10);inc(long_final,10);
copymemory(data,puntero,comprimido);
inc(data,comprimido);inc(long_final,comprimido);
freemem(puntero);
//TMS9918 longitud=81960
cantidad:=sizeof(TTMS99XX);
buffer[0]:=ord('T');
buffer[1]:=ord('M');
buffer[2]:=ord('S');
buffer[3]:=ord('R');
buffer[7]:=0;buffer[8]:=0;buffer[9]:=0;
getmem(puntero,cantidad);
ptemp:=pointer(TMS);
compress_zlib(ptemp,cantidad,pointer(puntero),comprimido);
buffer[4]:=(comprimido mod 65536) and $ff;
buffer[5]:=(comprimido mod 65536) shr 8;
buffer[6]:=comprimido div 65536;
copymemory(data,addr(buffer[0]),10);
inc(data,10);inc(long_final,10);
copymemory(data,puntero,comprimido);
inc(data,comprimido);inc(long_final,comprimido);
freemem(puntero);
//Registros Z80 longitud=68
cantidad:=sizeof(reg_z80);
buffer[0]:=ord('Z');
buffer[1]:=ord('8');
buffer[2]:=ord('0');
buffer[3]:=ord('R');
buffer[4]:=cantidad;
buffer[5]:=0;buffer[6]:=0;buffer[7]:=0;buffer[8]:=0;buffer[9]:=0;
copymemory(data,addr(buffer[0]),10);
inc(data,10);inc(long_final,10);
ptemp:=main_z80;
copymemory(data,ptemp,cantidad);
inc(long_final,cantidad);
write_file(nombre,datos_final,long_final);
freemem(datos_final);
end;

procedure coleco_interrupt(int:boolean);
begin
  if int then main_z80.pedir_nmi:=PULSE_LINE;
end;

end.
