unit mb88xx;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     main_engine,dialogs,sysutils,timer_engine;

type
     reg_mb88xx=record
        pc:byte; 	// Program Counter: 6 bits */
        pa:byte; 	// Page Address: 4 bits */
        sp:array[0..4-1] of word;	// Stack is 4*10 bit addresses deep, but we also use 3 top bits per address to store flags during irq */
        si:byte;		// Stack index: 2 bits */
        a:byte;		// Accumulator: 4 bits */
        x:byte;		// Index X: 4 bits */
        y:byte;		// Index Y: 4 bits */
        st:byte;		// State flag: 1 bit */
        zf:byte;		// Zero flag: 1 bit */
        cf:byte;		// Carry flag: 1 bit */
        vf:byte;		// Timer overflow flag: 1 bit */
        sf:byte;		// Serial Full/Empty flag: 1 bit */
        nf:byte;		// Interrupt flag: 1 bit */
        // Peripheral Control */
        pio:byte; // Peripheral enable bits: 8 bits */

        // Timer registers */
        th:byte;	// Timer High: 4 bits */
        tl:byte;	// Timer Low: 4 bits */
        tp:byte; // Timer Prescale: 6 bits? */
        ctr:byte; // current external counter value */

        // Serial registers */
        sb:byte;	// Serial buffer: 4 bits */
        sbcount:word;	// number of bits received */
///    emu_timer *serial;

        // PLA configuration */
        PLA_m:pbyte;

        // IRQ handling */
        pending_interrupt:byte;
        //    cpu_irq_callback irqcallback;
        rom:array[0..$7ff] of byte;
        ram:array[0..$7f] of byte;
        //      const address_space *io;
        numero_cpu:byte;
        tframes:single;
        opcode:boolean;
        contador:integer;
        halt:boolean;
     end;
     type_mb88xx_inport=function():byte;
     type_mb88xx_outport=procedure (valor:byte);
     type_mb88xx_inport_r=function(port:byte):byte;
     type_mb88xx_outport_r=procedure (port,valor:byte);
     llam_mb88xx=record
        port_k,port_p_r:type_mb88xx_inport;
        port_o,port_p_w:type_mb88xx_outport;
        port_r_r:type_mb88xx_inport_r;
        port_r_w:type_mb88xx_outport_r;
     end;
     preg_mb88xx=^reg_mb88xx;
const
  INT_CAUSE_SERIAL=01;
  INT_CAUSE_TIMER =02;
  INT_CAUSE_EXTERNAL=04;
  TIMER_PRESCALE=32;

procedure reset_mb88xx(r:preg_mb88xx);
procedure run_mb88xx(r:preg_mb88xx;ll:llam_mb88xx;maximo:dword);
procedure mb88xx_set_irq_line(r:preg_mb88xx;state:byte);

implementation

procedure mb88xx_set_irq_line(r:preg_mb88xx;state:byte);
begin
	// on falling edge trigger interrupt */
	if (((r.pio and $04)<>0) and (r.nf=0) and (state<>CLEAR_LINE)) then begin
		r.pending_interrupt:=r.pending_interrupt or INT_CAUSE_EXTERNAL;
	end;
  if state<>CLEAR_LINE then r.nf:=1
    else r.nf:=0;
end;

procedure reset_mb88xx(r:preg_mb88xx);
begin
  r.pc:=0;
	r.pa:=0;
	r.sp[0]:=0;
  r.sp[1]:=0;
  r.sp[2]:=0;
  r.sp[3]:=0;
	r.si:=0;
	r.a:=0;
	r.x:=0;
	r.y:=0;
	r.st:=1;	// start off with st=1 */
	r.zf:=0;
	r.cf:=0;
	r.vf:=0;
	r.sf:=0;
	r.nf:=0;
	r.pio:=0;
	r.th:=0;
	r.tl:=0;
	r.tp:=0;
	r.sb:=0;
	r.sbcount:= 0;
	r.pending_interrupt:= 0;
  r.halt:=false;
end;

procedure inc_pc(r:preg_mb88xx);inline;
begin
r.pc:=r.pc+1;
if r.pc=$40 then begin
  r.pc:=0;
  r.pa:=r.pa+1;
end;
end;

function get_pc(r:preg_mb88xx):word;inline;
begin
  get_pc:=(r.PA shl 6)+r.pc;
end;

function get_ea(r:preg_mb88xx):word;inline;
begin
  get_ea:=(r.x shl 4)+r.y;
end;

function pla(r:preg_mb88xx;inA,inB:byte):byte;inline;
var
  index:byte;
begin
	index:=((inB and 1) shl 4) or (inA and $0f);
//	if addr(r.PLA)<>nil then return cpustate->PLA[index];
  pla:=index;
end;

procedure update_pio_enable(r:preg_mb88xx;newpio:byte);inline;
begin
	// if the serial state has changed, configure the timer */
	if ((r.pio xor newpio) and $30)<>0 then begin
		if ((newpio and $30)=0) then //cpustate->serial->adjust(attotime::never);
		  else if ((newpio and $30)=$20) then //pustate->serial->adjust(attotime::from_hz(cpustate->device->clock() / SERIAL_PRESCALE), 0, attotime::from_hz(cpustate->device->clock() / SERIAL_PRESCALE));
		    else; //fatalerror("mb88xx: update_pio_enable set serial enable to unsupported value %02X\n", newpio & 0x30);
	end;
	r.pio:=newpio;
end;

function update_st_c(v:byte):byte;inline;
begin
  if (v and $10)<>0 then update_st_c:=0
    else update_st_c:=1;
end;

function update_st_z(v:byte):byte;inline;
begin
	 if (v=0) then update_st_z:=0
      else update_st_z:=1;
end;

function update_cf(v:byte):byte;inline;
begin
  if (v and $10)=0 then update_cf:=0
    else update_cf:=1;
end;

function update_zf(v:byte):byte;inline;
begin
  if (v<>0) then update_zf:=0
    else update_zf:=1;
end;

procedure increment_timer(r:preg_mb88xx);inline;
begin
	r.tl:=(r.tl+1) and $0f;
	if (r.tl=0) then begin
		r.th:=(r.th+1) and $0f;
		if (r.th=0) then begin
			r.vf:=1;
			r.pending_interrupt:=r.pending_interrupt or INT_CAUSE_TIMER;
		end;
	end;
end;

function update_pio(r:preg_mb88xx;cycles:byte):byte;
var
  estados:byte;
begin
estados:=0;
	// internal clock enable */
	if (r.pio and $80)<>0 then begin
		r.tp:=r.tp+cycles;
		while (r.tp>=TIMER_PRESCALE) do begin
			r.tp:=r.tp-TIMER_PRESCALE;
			increment_timer(r);
		end;
	end;
	// process pending interrupts */
	if (r.pending_interrupt and r.pio)<>0 then begin
		r.sp[r.si]:=get_pc(r);
		r.sp[r.si]:=r.sp[r.si] or (r.cf shl 15);
		r.sp[r.si]:=r.sp[r.si] or (r.zf shl 14);
		r.sp[r.si]:=r.sp[r.si] or (r.st shl 13);
		r.si:=(r.si+1) and 3;
		{/* the datasheet doesn't mention interrupt vectors but
        the Arabian MCU program expects the following }
		if (r.pending_interrupt and r.pio and INT_CAUSE_EXTERNAL)<>0 then begin
			// if we have a live external source, call the irqcallback
			//(*cpustate->irqcallback)(cpustate->device, 0);
			r.pc:=$02;
		end else if (r.pending_interrupt and r.pio and INT_CAUSE_TIMER)<>0 then begin
			r.pc:=$04;
		end else if (r.pending_interrupt and r.pio and INT_CAUSE_SERIAL)<>0 then begin
			r.pc:=$06;
		end;
		r.pa:=$00;
		r.st:=1;
		r.pending_interrupt:=0;
    estados:=3;
	end;
  update_pio:=estados;
end;

procedure run_mb88xx(r:preg_mb88xx;ll:llam_mb88xx;maximo:dword);
var
  instruccion,timming,arg,tempb:byte;
begin
if r.halt then begin
  r.contador:=maximo;
  exit;
end;
r.contador:=0;
while r.contador<maximo do begin
  timming:=1;
  instruccion:=r.rom[get_pc(r)];
  inc_pc(r);
  //Comprobar irq
  case instruccion of
    $00:r.st:=1; //nop
    $01:begin //outO
           if addr(ll.port_o)<>nil then ll.port_o(pla(r,r.a,r.cf));
           r.st:=1;
        end;
    $02:begin //outP
           if addr(ll.port_p_w)<>nil then ll.port_p_w(r.a);
           r.st:=1;
        end;
    $03:begin //outR
           arg:=r.y;
           if addr(ll.port_r_w)<>nil then ll.port_r_w(arg and 3,r.a);
				   r.st:=1;
        end;
    $04:begin //tay
           r.y:=r.a;
           r.st:=1;
        end;
    $05:begin  //tath
           r.th:=r.a;
           r.st:=1;
        end;
    $06:begin  //tatl
           r.tl:=r.a;
           r.st:=1;
        end;
    $07:begin  //tas
           r.sb:=r.a;
           r.st:=1;
        end;
    $08:begin //icy
           r.y:=r.y+1;
           r.st:=update_st_c(r.y);
				   r.y:=r.y and $0f;
				   r.zf:=update_zf(r.y);
        end;
    $09:begin //icm
           arg:=r.ram[get_ea(r)];
           arg:=arg+1;
           r.st:=update_st_c(arg);
				   arg:=arg and $0f;
				   r.zf:=update_zf(arg);
           r.ram[get_ea(r)]:=arg;
        end;
    $0a:begin //stic
           r.ram[get_ea(r)]:=r.a;
				   r.y:=r.y+1;
           r.st:=update_st_c(r.y);
				   r.y:=r.y and $0f;
				   r.zf:=update_zf(r.y);
        end;
    $0b:begin //x
           arg:=r.ram[get_ea(r)];
				   r.ram[get_ea(r)]:=r.a;
				   r.a:=arg;
           r.zf:=update_zf(r.a);
           r.st:=1;
        end;
    $0c:begin  // rol
				   r.a:=r.a shl 1;
           r.a:=r.a or r.cf;
				   r.st:=update_st_c(r.a);
				   r.cf:=r.st xor 1;
				   r.a:=r.a and $f;
				   r.zf:=update_zf(r.a);
        end;
    $0d:begin  //l
           r.a:=r.ram[get_ea(r)];
           r.zf:=update_zf(r.a);
           r.st:=1;
        end;
    $0e:begin //adc
           arg:=r.ram[get_ea(r)];
           arg:=arg+r.a+r.cf;
           r.st:=update_st_c(arg);
				   r.cf:=r.st xor 1;
				   r.a:=arg and $0f;
           r.zf:=update_zf(r.a);
        end;
    $12:begin //inK
           if addr(ll.port_k)<>nil then r.a:=ll.port_k and $f
            else r.a:=$f;
           r.zf:=update_zf(r.a);
           r.st:=1;
        end;
    $13:begin // inR
           if addr(ll.port_r_r)<>nil then r.a:=(ll.port_r_r(r.y and 3)) and $f
            else r.a:=$f;
           r.zf:=update_zf(r.a);
           r.st:=1;
				end;
    $14:begin //tya
          r.a:=r.y;
				  r.zf:=update_zf(r.a);
          r.st:=1;
        end;
    $15:begin //ttha
          r.a:=r.th;
				  r.zf:=update_zf(r.a);
          r.st:=1;
        end;
    $16:begin //ttla
          r.a:=r.tl;
				  r.zf:=update_zf(r.a);
          r.st:=1;
        end;
    $17:begin //tsa
          r.a:=r.sb;
          r.zf:=update_zf(r.a);
          r.st:=1;
        end;
    $18:begin //dcy
          r.y:=r.y-1;
          r.st:=update_st_c(r.y);
          r.y:=r.y and $f;
        end;
    $19:begin //dcm
          arg:=r.ram[get_ea(r)];
          arg:=arg-1;
          r.st:=update_st_c(arg);
          arg:=arg and $f;
          r.zf:=update_zf(arg);
          r.ram[get_ea(r)]:=arg;
        end;
    $1b:begin //xx
           arg:=r.x;
           r.x:=r.a;
           r.a:=arg;
           r.zf:=update_zf(r.a);
				   r.st:=1;
        end;
    $1c:begin  // ror
           r.a:=r.a or (r.cf shl 4);
           r.st:=update_st_c(r.a shl 4);
           r.cf:=r.st xor 1;
           r.a:=(r.a shr 1) and $f;
           r.zf:=update_zf(r.a);
        end;
    $1d:begin //st
           r.ram[get_ea(r)]:=r.a;
				   r.st:=1;
        end;
    $1e:begin //sbc
           arg:=r.ram[get_ea(r)];
           arg:=arg-r.a-r.cf;
           r.st:=update_st_c(arg);
           r.cf:=r.st xor 1;
           r.a:=arg and $0f;
           r.zf:=update_zf(r.a);
        end;
    $1f:begin   //or
           r.a:=r.a or r.ram[get_ea(r)];
           r.zf:=update_zf(r.a);
           r.st:=r.zf xor 1;
        end;
    $21:begin  //setc
           r.cf:=1;
           r.st:=1;
        end;
    $23:begin  //rstc
           r.cf:=0;
           r.st:=1;
        end;
    $24:begin //tstr
           if addr(ll.port_r_r)<>nil then arg:=ll.port_r_r(r.y div 4) and $f
            else arg:=$f;
				   if (arg and (1 shl (r.y mod 4)))<>0 then r.st:=0
            else r.st:=1;
        end;
    $25:r.st:=r.nf xor 1; // tsti
    $28:r.st:=r.cf xor 1; //tstc
    $2c:begin  //rts
           r.si:=(r.si-1) and $3;
           r.pc:=r.sp[r.si] and $3f;
           r.pa:=(r.sp[r.si] shr 6) and $1f;
           r.st:=1;
        end;
    $2e:begin //c
           arg:=r.ram[get_ea(r)];
           arg:=arg-r.a;
           r.cf:=update_cf(arg);
				   arg:=arg and $0f;
           r.st:=update_st_z(arg);
           r.zf:=r.st xor 1;
        end;
    $30..$33:begin //sbit
           arg:=r.ram[get_ea(r)];
           r.ram[get_ea(r)]:=arg or (1 shl (instruccion and 3));
           r.st:=1;
        end;
    $34..$37:begin //rbit
           arg:=r.ram[get_ea(r)];
           r.ram[get_ea(r)]:=arg and not(1 shl (instruccion and 3));
           r.st:=1;
        end;
    $38..$3b:begin //tbit
           arg:=r.ram[get_ea(r)];
           if (arg and (1 shl (instruccion and $3)))<>0 then r.st:=0
            else r.st:=1;
        end;
    $3c:begin //rti
           r.si:=(r.si-1) and 3;
				   r.pc:=r.sp[r.si] and $3f;
				   r.pa:=(r.sp[r.si] shr 6) and $1f;
				   r.st:=(r.sp[r.si] shr 13) and 1;
				   r.zf:=(r.sp[r.si] shr 14) and 1;
				   r.cf:=(r.sp[r.si] shr 15) and 1;
        end;
    $3d:begin //jpa imm
           r.pa:=r.rom[get_pc(r)] and $1f;
				   r.pc:=r.a*4;
				   timming:=2;
				   r.st:=1;
        end;
    $3e:begin //en imm
           update_pio_enable(r, r.pio or r.rom[get_pc(r)]);
				   inc_pc(r);
				   timming:=2;
				   r.st:=1;
        end;
    $3f:begin  //dis
           update_pio_enable(r, r.pio and not(r.rom[get_pc(r)]));
				   inc_pc(r);
				   timming:=2;
				   r.st:=1;
        end;
    $40..$43:begin  //setD
           if addr(ll.port_r_r)<>nil then arg:=ll.port_r_r(0) and $f
            else arg:=$f;
           arg:=arg or (1 shl (instruccion and 3));
           if addr(ll.port_r_w)<>nil then ll.port_r_w(0,arg);
           r.st:=1;
        end;
    $44..$47:begin  //rstD
           if addr(ll.port_r_r)<>nil then arg:=ll.port_r_r(0) and $f
            else arg:=$f;
           arg:=arg and (not (1 shl (instruccion and 3)));
           if addr(ll.port_r_w)<>nil then ll.port_r_w(0,arg);
           r.st:=1;
        end;
    $4c..$4f:if (r.a and (1 shl (instruccion and 3)))<>0 then r.st:=0
                else r.st:=1; //tba
    $50..$53:begin  //xd
           arg:=r.ram[instruccion and $3];
           r.ram[instruccion and $3]:=r.a;
				   r.a:=arg;
				   r.zf:=update_zf(r.a);
				   r.st:=1;
        end;
    $54..$57:begin //xyd
           arg:=r.ram[(instruccion and 3)+4];
           r.ram[(instruccion and 3)+4]:=r.y;
           r.y:=arg;
           r.zf:=update_zf(r.y);
           r.st:=1;
        end;
    $58..$5f:begin //lxi
           r.x:=instruccion and $7;
           r.zf:=update_zf(r.x);
				   r.st:=1;
        end;
    $60..$67:begin //call imm
           arg:=r.rom[get_pc(r)];
				   inc_pc(r);
				   timming:=2;
           if r.st<>0 then begin
              r.sp[r.si]:=get_pc(r);
              r.si:=(r.si+1) and $3;
              r.pc:=arg and $3f;
              r.pa:=((instruccion and 7) shl 2) or (arg shr 6);
				   end;
				   r.st:=1;
        end;
    $68..$6f:begin //jpl imm
           arg:=r.rom[get_pc(r)];
           inc_pc(r);
           timming:=2;
				   if r.st<>0 then begin
					    r.pc:=arg and $3f;
					    r.pa:=((instruccion and 7) shl 2) or (arg shr 6);
				   end;
				   r.st:=1;
        end;
    $70..$7f:begin //ai
           arg:=instruccion and $0f;
				   arg:=arg+r.a;
           r.st:=update_st_c(arg);
           r.cf:=r.st xor 1;
           r.a:=arg and $f;
           r.zf:=update_zf(r.a);
        end;
    $80..$8f:begin //lyi
           r.y:=instruccion and $0f;
           r.zf:=update_zf(r.y);
				   r.st:=1;
        end;
    $90..$9f:begin //li
           r.a:=instruccion and $0f;
           r.zf:=update_zf(r.a);
				   r.st:=1;
        end;
    $a0..$af:begin //cyi
           arg:=(instruccion and $0f)-r.y;
           r.cf:=update_cf(arg);
           arg:=arg and $f;
           r.st:=update_st_z(arg);
           r.zf:=r.st xor 1;
        end;
    $b0..$bf:begin //ci
           arg:=(instruccion and $0f)-r.a;
           r.cf:=update_cf(arg);
				   arg:=arg and $0f;
           r.st:=update_st_z(arg);
				   r.zf:=r.st xor 1;
        end;
    $c0..$ff:begin //jmp
           if r.st<>0 then r.pc:=instruccion and $3f;
				   r.st:=1;
        end;
    else MessageDlg('Instruccion MB88XX '+inttohex(instruccion,2)+' desconocida. PC='+inttohex(get_pc(r),10), mtInformation,[mbOk], 0);
  end;
tempb:=timming+update_pio(r,timming);
r.contador:=r.contador+tempb;
update_timer(tempb,r.numero_cpu);
end;
end;

end.
