unit tms32010;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     main_engine,dialogs,sysutils;

type
        band_tms32010=record
                ov,ovm,int:boolean;
                dp_reg:boolean;
                arp_reg:byte;
        end;
        reg_tms32010=record
                rom:array[0..$fff] of word;
                ram:array[0..$ff] of word;
                pc:parejas;
                sp:array[0..3] of word;
                str:band_tms32010;
                acc,oldacc,alu:dparejas;
                preg,treg:dword;
                ar:array[0..1] of word;
                memaccess:word;
                contador:longint;
                numero_cpu:byte;
                tframes:single;
                opcode,enabled:boolean;
                pedir_int:byte;
        end;
        type_inport=function():word; 
        type_outport=procedure (valor:word); 
        llam_tms32010 = record
                      in_bio:function:boolean; 
                      in_port:array[0..7] of type_inport;
                      out_port:array[0..7] of type_outport;
                    end;
        preg_tms32010=^reg_tms32010;
const

  ciclos_tms:array[0..$ff] of byte=(
    	 // 0 1 2 3 4 5 6 7 8 9 A B C D E F
          1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
          1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
          1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
          1,1,0,0,0,0,0,0,1,1,0,0,0,0,0,0,
          2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
          1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,
          1,1,1,1,1,1,1,3,1,1,1,1,1,1,1,1,
          1,1,0,0,0,0,0,0,1,1,1,1,1,3,1,0,
          1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
          1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
          0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          0,0,0,0,2,2,2,0,2,2,2,2,2,2,2,2);

procedure run_tms32010(r:preg_tms32010;ll:llam_tms32010;maximo:dword);
procedure reset_tms32010(r:preg_tms32010);

implementation

procedure reset_tms32010(r:preg_tms32010);
begin
r.opcode:=false;
r.pc.w:=0;
r.acc.l:=0;
r.sp[0]:=0;
r.sp[1]:=0;
r.sp[2]:=0;
r.sp[3]:=0;
r.pedir_int:=CLEAR_LINE;
r.str.arp_reg:=0;
r.enabled:=true;
end;

procedure push_stack(r:preg_tms32010;valor:word);inline;
begin
r.sp[0]:=r.sp[1];
r.sp[1]:=r.sp[2];
r.sp[2]:=r.sp[3];
r.sp[3]:=valor;
end;

function pop_stack(r:preg_tms32010):word;inline;
var
  ret:word;
begin
  ret:=r.sp[3];
	r.sp[3]:=r.sp[2];
	r.sp[2]:=r.sp[1];
	r.sp[1]:=r.sp[0];
	pop_stack:=ret;
end;

procedure calculate_add_overflow(r:preg_tms32010;addval:longint);inline;
var
  temp:longint;
begin
  temp:=longint(not(r.oldacc.l xor addval) and (r.oldacc.l xor r.acc.l));
  if temp<0 then begin
    r.str.ov:=true;
    if r.str.ovm then begin
      if longint(r.oldacc.l)<0 then r.acc.l:=$80000000
        else r.acc.l:=$7fffffff;
    end;
	end;
end;

procedure calculate_sub_overflow(r:preg_tms32010;subval:longint);inline;
var
  temp:longint;
begin
  temp:=longint((r.oldacc.l xor subval) and (r.oldacc.l xor r.acc.l));
  if temp<0 then begin
    r.str.ov:=true;
    if r.str.ovm then begin
      if longint(r.oldacc.l)<0 then r.acc.l:=$80000000
        else r.acc.l:=$7fffffff;
    end;
	end;
end;

procedure update_ar(r:preg_tms32010;valor:byte);inline;
var
  tmp:word;
begin
	if (valor and $30)<>0 then begin
		tmp:=r.ar[r.str.arp_reg];
		if (valor and $20)<>0 then tmp:=tmp+1;
		if (valor and $10)<>0 then tmp:=tmp-1;
    r.ar[r.str.arp_reg]:=(r.ar[r.str.arp_reg] and $fe00) or (tmp and $01ff);
	end;
end;

procedure update_arp(r:preg_tms32010;valor:byte);inline;
begin
	if (valor and $08)=0 then r.str.arp_reg:=(valor and $01);
end;

procedure putdata(r:preg_tms32010;valor:word;instruccion:byte);inline;
begin
  if (instruccion and $80)<>0 then begin //ind
    r.memaccess:=r.ar[r.str.arp_reg];
    update_ar(r,instruccion);
		update_arp(r,instruccion);
  end	else r.memaccess:=instruccion and $7f;
  r.ram[r.memaccess]:=valor;
end;

procedure putdata_sar(r:preg_tms32010;ar:byte;instruccion:byte);inline;
begin
  if (instruccion and $80)<>0 then begin //ind
    r.memaccess:=r.ar[r.str.arp_reg];
    update_ar(r,instruccion);
		update_arp(r,instruccion);
  end	else r.memaccess:=instruccion and $7f;
  r.ram[r.memaccess]:=r.ar[ar];
end;

procedure getdata(r:preg_tms32010;instruccion,shift:byte;signext:boolean);inline;
begin
  if (instruccion and $80)<>0 then r.memaccess:=r.ar[r.str.arp_reg]
    else r.memaccess:=instruccion and $7f;
	r.alu.l:=word(r.ram[r.memaccess]);
	if signext then r.alu.l:=smallint(r.alu.l);
  r.alu.l:=r.alu.l shl shift;
  if (instruccion and $80)<>0 then begin
    update_ar(r,instruccion);
		update_arp(r,instruccion);
  end
end;

function ext_irq(r:preg_tms32010):byte;inline;
var
  ret:byte;
begin
ret:=0;
if not(r.str.int) then begin
    r.pedir_int:=CLEAR_LINE;
    r.str.int:=true;
		push_stack(r,r.pc.w);
		r.pc.w:=$0002;
		ret:=3;	// 3 cycles used due to PUSH and DINT operation ? */
end;
ext_irq:=ret;
end;

procedure run_tms32010(r:preg_tms32010;ll:llam_tms32010;maximo:dword);
var
  estados_demas:byte;
  instruccion:parejas;
begin
r.contador:=0;
while r.contador<maximo do begin
if not(r.enabled)  then begin
  r.contador:=maximo;
  exit;
end;
estados_demas:=0;
//comprobar irq's
if (r.pedir_int<>CLEAR_LINE) then begin
  if ((instruccion.h<>$6d) and ((instruccion.h and $e0)<>$80) and (instruccion.w<>$7f82)) then estados_demas:=ext_irq(r);
end;
r.opcode:=true;
instruccion.w:=r.rom[r.pc.w];
r.pc.w:=r.pc.w+1;
r.opcode:=false;
case instruccion.h of
  $00..$0f:begin  //add_sh
              r.oldacc.l:=r.acc.l;
              getdata(r,instruccion.l,instruccion.h and $f,true);
              r.acc.l:=r.acc.l+r.alu.l;
              calculate_add_overflow(r,r.alu.l);
           end;
  $10..$1f:begin //sub_sh
              r.oldacc.l:=r.acc.l;
              getdata(r,instruccion.l,instruccion.h and $f,true);
              r.acc.l:=r.acc.l-r.alu.l;
              calculate_sub_overflow(r,r.alu.l);
           end;
  $20..$2f:begin //lac_sh
              getdata(r,instruccion.l,instruccion.h and $f,true);
	            r.acc.l:=r.alu.l;
           end;
  $30:putdata_sar(r,0,instruccion.l); //sar_ar0
  $31:putdata_sar(r,1,instruccion.l); //sar_ar1
  $32..$37,$3a..$3f,$51..$57,$72..$77,$a0..$f3,$f7:MessageDlg('Intruccion ilegal!!! DSP '+inttostr(r.pc.w), mtInformation,[mbOk], 0);
  $38:begin  //lar ar0
          getdata(r,instruccion.l,0,false);
          r.ar[0]:=r.alu.wl;
      end;
  $39:begin  //lar ar1
          getdata(r,instruccion.l,0,false);
          r.ar[1]:=r.alu.wl;
      end;
  $40..$47:begin  //in_p
              r.alu.l:=ll.in_port[instruccion.h and $7];
              putdata(r,r.alu.wl,instruccion.l);
           end;
  $48..$4f:begin  //out_p
              getdata(r,instruccion.l,0,false);
              if addr(ll.out_port[instruccion.h and 7])<>nil then ll.out_port[instruccion.h and 7](r.alu.wl)
                else MessageDlg('OUT sin funcion! '+inttostr(instruccion.h and 7), mtInformation,[mbOk], 0);
      end;
  $50:putdata(r,r.acc.wl,instruccion.l);  //sacl
  $58..$5f:begin //sach_sh
              r.alu.l:=(r.acc.l shl (instruccion.h and $7));
              putdata(r,r.alu.wh,instruccion.l);
           end;
  $60:begin  //addh
        r.oldacc.l:=r.acc.l;
        getdata(r,instruccion.l,0,false);
        r.acc.wh:=r.acc.wh+r.alu.wl;
	      if (smallint((not(r.oldacc.wh xor r.ALU.wh)) and (r.oldacc.wh xor r.acc.wh)) < 0) then begin
      		r.str.ov:=true;
      		if r.str.ovm then begin
            if (smallint(r.oldacc.wh)<0) then r.ACC.wh:=$8000
              else r.ACC.wh:=$7fff;
          end;
	      end;
      end;
  $61:begin //adds
        r.oldacc.l:=r.acc.l;
        getdata(r,instruccion.l,0,false);
        r.acc.l:=r.acc.l+r.alu.l;
        calculate_add_overflow(r,r.alu.l);
      end;
  $62:begin  //subh
        r.oldacc.l:=r.acc.l;
        getdata(r,instruccion.l,16,false);
        r.acc.l:=r.acc.l-r.alu.l;
        calculate_sub_overflow(r,r.alu.l);
      end;
  $63:begin //subs
        r.oldacc.l:=r.acc.l;
        getdata(r,instruccion.l,0,false);
        r.acc.l:=r.acc.l-r.alu.l;
        calculate_sub_overflow(r,r.alu.l);
      end;
  $65:begin  //zalh
        getdata(r,instruccion.l,0,false);
        r.acc.wh:=r.alu.wl;
        r.acc.wl:=$0000;
      end;
  $66:begin //zals
        getdata(r,instruccion.l,0,false);
        r.acc.wl:=r.alu.wl;
        r.acc.wh:=$0000;
      end;
  $67:begin //tblr
	      r.alu.l:=r.rom[r.acc.wl];
        putdata(r,r.alu.wl,instruccion.l);
        r.sp[0]:=r.sp[1];
      end;
  $68:if (instruccion.l and $80)<>0 then begin //larp_mar
		    update_ar(r,instruccion.l);
		    update_arp(r,instruccion.l);
	    end;
  $69:begin //dmov
         getdata(r,instruccion.l,0,false);
         r.ram[r.memaccess+1]:=r.alu.wl;
      end;
  $6a:begin //lt
        getdata(r,instruccion.l,0,false);
        r.treg:=r.alu.l;
      end;
  $6d:begin //mpy
        getdata(r,instruccion.l,0,false);
        r.preg:=smallint(r.alu.wl)*smallint(r.treg);
	      if (r.preg=$40000000) then r.preg:=$c0000000;
      end;
  $6e:r.str.dp_reg:=(instruccion.l and 1)<>0; //ldpk
  $70:r.ar[0]:=instruccion.l; //lark0
  $71:r.ar[1]:=instruccion.l; //lark1
  $78:begin  //xor
        getdata(r,instruccion.l,0,false);
        r.acc.l:=r.acc.l xor r.alu.l;
      end;
  $79:begin //and
        getdata(r,instruccion.l,0,false);
        r.acc.l:=r.acc.l and r.alu.l;
      end;
  $7a:begin  //or
        getdata(r,instruccion.l,0,false);
        r.acc.l:=r.acc.l or r.alu.l;
      end;
  $7e:r.acc.l:=instruccion.l; //lack
  $7f:begin //el $7f es compuesto
        estados_demas:=estados_demas+1;
        case instruccion.l of
          $80:; //nop
          $81:r.str.int:=true; //dint
          $82:r.str.int:=false;  //eint
          $88:if (longint(r.acc.l)<0) then begin //abst
                r.acc.l:=-r.acc.l;
		            if (r.str.ovm and (r.acc.l=$80000000)) then r.acc.l:=r.acc.l-1;
              end;
          $89:r.acc.l:=0; //zac
          $8a:r.str.ovm:=false;  //rovm
          $8c:begin //cala
                push_stack(r,r.pc.w);
                r.pc.w:=r.acc.wl;
              end;
          $8d:r.pc.w:=pop_stack(r); //ret
          $8e:r.acc.l:=r.preg; //pac
            else MessageDlg('Intruccion desconocida $7f DSP '+inttostr(r.pc.w-1), mtInformation,[mbOk], 0);
        end;
      end;
  $80..$9f:r.preg:=smallint(r.treg)*(smallint(instruccion.l shl 3) shr 3); //mpyk
  $f4:begin //banz
        if (r.ar[r.str.arp_reg] and $01ff)<>0 then begin
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
        end else r.pc.w:=r.pc.w+1;
	      r.alu.wl:=r.ar[r.str.arp_reg];
        r.alu.wl:=r.alu.wl-1;
        r.ar[r.str.arp_reg]:=(r.ar[r.str.arp_reg] and $fe00) or (r.alu.wl and $1ff);
      end;
  $f5:if r.str.ov then begin  //bv
          r.pc.w:=r.rom[r.pc.w];
          r.str.ov:=false;
          estados_demas:=estados_demas+1;
	    end	else r.pc.w:=r.pc.w+1;
  $f6:if ll.in_bio then begin  //bioz
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  $f8:begin //call
        r.pc.w:=r.pc.w+1;
	      push_stack(r,r.pc.w);
        r.pc.w:=r.rom[r.pc.w-1];
      end;
  $f9:r.pc.w:=r.rom[r.pc.w]; //br
  $fa:if longint(r.acc.l)<0 then begin //blz
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  $fb:if (longint(r.acc.l)<=0) then begin //blez
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  $fc:if longint(r.acc.l)>0 then begin  //bgz
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
        end else r.pc.w:=r.pc.w+1;
  $fd:if longint(r.acc.l)>=0 then begin //bgez
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  $fe:if (r.acc.l<>0) then begin  //bnz
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  $ff:if (r.acc.l=0) then begin  //bz
          r.pc.w:=r.rom[r.pc.w];
          estados_demas:=estados_demas+1;
      end else r.pc.w:=r.pc.w+1;
  else MessageDlg('Intruccion desconocida DSP '+inttohex(r.pc.w-1,10), mtInformation,[mbOk], 0);
end; //del case
r.contador:=r.contador+ciclos_tms[instruccion.h]+estados_demas;
//update_timer(ciclos_tms[instruccion.h]+estados_demas,r.numero_cpu);
end; //del while!
end;

end.
