unit pal_engine;

interface
uses sdl;

const
    MAX_NETS=3;
    MAX_RES_PER_NET=18;
    max_colors=$fff;
    set_trans_color=$1;
    max_colores=$8000;
type
  tcolor=record
                r,g,b:byte;
         end;
  tpaleta=array[0..max_colores] of tcolor;
var
  paleta:array[0..max_colores] of word;
  buffer_paleta:array[0..max_colores*2] of byte;

procedure compute_resistor_weights(
	minval,maxval:integer;scaler:double;
	count_1:integer;resistances_1:pinteger;weights_1:pdouble;pulldown_1,pullup_1:integer;
	count_2:integer;resistances_2:pinteger;weights_2:pdouble;pulldown_2,pullup_2:integer;
	count_3:integer;resistances_3:pinteger;weights_3:pdouble;pulldown_3,pullup_3:integer);
function combine_2_weights(tab:pdouble;w0,w1:integer):integer;
function combine_3_weights(tab:pdouble;w0,w1,w2:integer):integer;
function combine_4_weights(tab:pdouble;w0,w1,w2,w3:integer):integer;
function combine_6_weights(tab:pdouble;w0,w1,w2,w3,w4,w5:integer):integer;
function pal2bit(bits:byte):byte;
function pal3bit(bits:byte):byte;
function pal4bit(bits:byte):byte;
function pal5bit(bits:byte):byte;
//Palette functions
procedure set_pal(ppaleta:tpaleta;total_colors:word);
procedure set_pal_color(pcolor:tcolor;ppaleta:pword);

implementation
uses main_engine;

procedure compute_resistor_weights(
	minval,maxval:integer;scaler:double;
	count_1:integer;resistances_1:pinteger;weights_1:pdouble;pulldown_1,pullup_1:integer;
	count_2:integer;resistances_2:pinteger;weights_2:pdouble;pulldown_2,pullup_2:integer;
	count_3:integer;resistances_3:pinteger;weights_3:pdouble;pulldown_3,pullup_3:integer);
var
	networks_no:integer;
	rescount:array[0..(MAX_NETS-1)] of integer;		// number of resistors in each of the nets */
	r:array[0..(MAX_NETS-1),0..(MAX_RES_PER_NET-1)] of double;		// resistances */
	w:array[0..(MAX_NETS-1),0..(MAX_RES_PER_NET-1)] of double;		// calulated weights */
	ws:array[0..(MAX_NETS-1),0..(MAX_RES_PER_NET-1)] of double;	// calulated, scaled weights */
	r_pd:array[0..(MAX_NETS-1)] of integer;			// pulldown resistances */
	r_pu:array[0..(MAX_NETS-1)] of integer;			// pullup resistances */

	max_out:array[0..(MAX_NETS-1)] of double;
	out_:array[0..(MAX_NETS-1)] of pdouble;

	i,j,n:integer;
	scale:double;
	max:double;
  count, pd, pu:integer;
	resistances:pinteger;
	weights:pdouble;
  R0,R1,Vout,dst,sum:double;
  out2_:pdouble;
begin
	// parse input parameters */
	networks_no:=0;
	for n:=0 to (MAX_NETS-1) do begin
		case n of
		  0:begin
				  count:=count_1;
				  resistances:=resistances_1;
				  weights:=weights_1;
				  pd:=pulldown_1;
				  pu:=pullup_1;
				end;
		  1:begin
				  count:=count_2;
				  resistances:=resistances_2;
				  weights:=weights_2;
				  pd:=pulldown_2;
				  pu:=pullup_2;
				end;
		  else begin
				  count:=count_3;
				  resistances:=resistances_3;
				  weights:=weights_3;
				  pd:=pulldown_3;
				  pu:=pullup_3;
        end;
		end;
		// parameters validity check */
	{	if (count>MAX_RES_PER_NET) then
			fatalerror("compute_resistor_weights(): too many resistors in net #%i. The maximum allowed is %i, the number requested was: %i\n",n, MAX_RES_PER_NET, count);
   }

		if (count > 0) then begin
			rescount[networks_no]:=count;
			for i:=0 to (count-1) do begin
				r[networks_no,i]:=1.0*pintegerarray(resistances)^[i];
			end;
			out_[networks_no]:=weights;
			r_pd[networks_no]:=pd;
			r_pu[networks_no]:=pu;
			inc(networks_no);
		end;
	end;
 {	if (networks_no<1)
		fatalerror("compute_resistor_weights(): no input data\n");
  }
	// calculate outputs for all given networks */
	for i:=0 to (networks_no-1) do begin
		// of n resistors */
		for n:=0 to (rescount[i]-1) do begin
      if (r_pd[i]=0) then R0:=1.0/1e12
        else R0:=1.0/r_pd[i];
      if (r_pu[i]=0) then R1:=1.0/1e12
        else R1:=1.0/r_pu[i];
			for j:=0 to (rescount[i]-1) do begin
				if (j=n) then begin	// only one resistance in the network connected to Vcc */
					if (r[i,j]<>0.0) then R1:=R1+(1.0/r[i,j]);
				end else begin
					if (r[i,j]<>0.0) then R0:=R0+(1.0/r[i,j]);
        end;
			end;

			// now determine the voltage */
			R0:=1.0/R0;
			R1:=1.0/R1;
			Vout:=(maxval-minval)*R0/(R1+R0)+minval;

			// and convert it to a destination value */
      if (Vout<minval) then begin
        dst:=minval
      end else begin
        if (Vout > maxval) then dst:=maxval
          else dst:=Vout;
      end;
			w[i,n]:=dst;
    end;
	 end;

	// calculate maximum outputs for all given networks
	j:=0;
	max:=0.0;
	for i:=0 to (networks_no-1) do begin
		sum:=0.0;
		// of n resistors */
		for n:=0 to (rescount[i]-1) do
			sum:=sum+w[i,n];	// maximum output, ie when each resistance is connected to Vcc */
		max_out[i]:=sum;
		if (max<sum) then begin
			max:=sum;
			j:=i;
		end;
	end;
	if (scaler<0.0)	then // use autoscale ? */
		// calculate the output scaler according to the network with the greatest output */
		scale:=maxval/max_out[j]
	else				// use scaler provided on entry */
		scale:=scaler;

	// calculate scaled output and fill the output table(s)*/
	for i:=0 to (networks_no-1) do begin
		for n:=0 to (rescount[i]-1) do begin
			ws[i,n]:=w[i,n]*scale;	// scale the result */
      out2_:=out_[i];
      inc(out2_,n);
			out2_^:=ws[i,n];		// fill the output table */
		end;
	end;
end;

function combine_2_weights(tab:pdouble;w0,w1:integer):integer;
var
  res:double;
  ptemp:pdouble;
begin
  ptemp:=tab;
  res:=ptemp^*w0;
  inc(ptemp);
  res:=res+ptemp^*w1+0.5;
  if res>255 then res:=255;
  combine_2_weights:=trunc(res);
end;

function combine_3_weights(tab:pdouble;w0,w1,w2:integer):integer;
var
  res:double;
  ptemp:pdouble;
begin
  ptemp:=tab;
  res:=ptemp^*w0;
  inc(ptemp);
  res:=res+ptemp^*w1;
  inc(ptemp);
  res:=res+ptemp^*w2+0.5;
  if res>255 then res:=255;
  combine_3_weights:=trunc(res);
end;

function combine_4_weights(tab:pdouble;w0,w1,w2,w3:integer):integer;
var
  res:double;
  ptemp:pdouble;
begin
  ptemp:=tab;
  res:=ptemp^*w0;
  inc(ptemp);
  res:=res+ptemp^*w1;
  inc(ptemp);
  res:=res+ptemp^*w2;
  inc(ptemp);
  res:=(res+ptemp^*w3)+0.5;
  if res>255 then res:=255;
  combine_4_weights:=trunc(res);
end;

function combine_6_weights(tab:pdouble;w0,w1,w2,w3,w4,w5:integer):integer;
var
  res:double;
  ptemp:pdouble;
begin
  ptemp:=tab;
  res:=ptemp^*w0;
  inc(ptemp);
  res:=res+ptemp^*w1;
  inc(ptemp);
  res:=res+ptemp^*w2;
  inc(ptemp);
  res:=res+ptemp^*w3;
  inc(ptemp);
  res:=res+ptemp^*w4;
  inc(ptemp);
  res:=(res+ptemp^*w5)+0.5;
  if res>255 then res:=255;
  combine_6_weights:=trunc(res);
end;

function pal2bit(bits:byte):byte;inline;
begin
	bits:=bits and 3;
	pal2bit:=(bits shl 6) or (bits shl 4) or (bits shl 2) or bits;
end;

function pal3bit(bits:byte):byte;inline;
begin
	bits:=bits and 7;
	pal3bit:=(bits shl 5) or (bits shl 2) or (bits shr 1);
end;

function pal4bit(bits:byte):byte;inline;
begin
	bits:=bits and $f;
	pal4bit:=(bits shl 4) or bits;
end;

function pal5bit(bits:byte):byte;inline;
begin
	bits:=bits and $1f;
	pal5bit:=(bits shl 3) or (bits shr 2);
end;

//Palette functions
procedure set_pal(ppaleta:tpaleta;total_colors:word);inline;
var
  colors:word;
begin
for colors:=0 to total_colors-1 do
        paleta[colors]:=SDL_MapRGB(pantalla[0].format, ppaleta[colors].r, ppaleta[colors].g, ppaleta[colors].b);
end;

procedure set_pal_color(pcolor:tcolor;ppaleta:pword);inline;
begin
ppaleta^:=SDL_MapRGB(pantalla[0].format, pcolor.r, pcolor.g, pcolor.b);
end;

end.
