unit dac;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses sound_engine;

type
 dac_info=record
	  output:integer;
	  UnsignedVolTable:array[0..256-1] of word;
	  SignedVolTable:array[0..256-1] of smallint;
    tsample:byte;
  end;
  pdac_info=^dac_info;
var
  chip_dac:array[0..2] of pdac_info;

procedure dac_init(numero:byte);
procedure DAC_update(numero:byte);
procedure DAC_signed_data_w(numero,data:byte);
procedure DAC_data_w(numero,data:byte);
procedure dac_reset(numero:byte);
procedure dac_close(numero:byte);

implementation

procedure DAC_update(numero:byte);
begin
  tsample[chip_dac[numero].tsample,sound_status.posicion_sonido]:=chip_dac[numero].output;
  if sound_status.stereo then tsample[chip_dac[numero].tsample,sound_status.posicion_sonido+1]:=chip_dac[numero].output;
end;

procedure DAC_data_w(numero,data:byte);
begin
	chip_dac[numero].output:=chip_dac[numero].UnsignedVolTable[data];
end;

procedure DAC_signed_data_w(numero,data:byte);
begin
	chip_dac[numero].output:=chip_dac[numero].SignedVolTable[data];
end;

procedure DAC_build_voltable(numero:byte);
var
  i:integer;
begin
	// build volume table (linear) */
	for i:=0 to 255 do begin
		chip_dac[numero].UnsignedVolTable[i]:=i*$101 shr 1;	// range      0..32767 */
		chip_dac[numero].SignedVolTable[i]:=i*$101-$8000;	// range -32768..32767 */
	end;
end;

procedure dac_reset(numero:byte);
begin
  chip_dac[numero].output:=0;
end;

procedure dac_init(numero:byte);
begin
  if chip_dac[numero]=nil then getmem(chip_dac[numero],sizeof(dac_info));
	fillchar(chip_dac[numero]^,sizeof(dac_info),0);
	DAC_build_voltable(numero);
  dac_reset(numero);
  chip_dac[numero].tsample:=init_channel;
end;

procedure dac_close(numero:byte);
begin
  if chip_dac[numero]<>nil then begin
    freemem(chip_dac[numero]);
    chip_dac[numero]:=nil;
  end;
end;


end.
