unit konami_snd;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses ay_8910,cpu_engine,sound_engine;

type
  tkonami_call=function(valor:byte):byte;
  tksound=record
    sound_latch,frame,tipo:byte;
    last_cycles,clock:integer;
    call:tkonami_call;
  end;
const
  ONE_AY8910=0;
  TWO_AY8910=1;

var
  konami_sound:tksound;

function konamisnd_porta:byte;
function konamisnd_portb:byte;
function konamisnd_getbyte(direccion:word):Byte;
procedure konamisnd_putbyte(direccion:word;valor:byte);
procedure konamisnd_2ay;
procedure konamisnd_1ay;
procedure konamisnd_init(amp,cpu,ntipo:byte;clock:integer;call:tkonami_call);
procedure konamisnd_reset;
procedure konamisnd_close;

implementation
uses principal;

function konamisnd_porta:byte;
begin
  konamisnd_porta:=konami_sound.sound_latch;
end;

function konamisnd_portb:byte;
const reloj:array[0..9] of byte=(0,$10,$20,$30,$40,$90,$a0,$b0,$a0,$d0);
var
  current_totalcycles:integer;
  res:byte;
begin
current_totalcycles:=snd_z80.contador+round(konami_sound.frame*snd_z80.tframes);
konami_sound.clock:=(konami_sound.clock+(current_totalcycles-konami_sound.last_cycles));
konami_sound.last_cycles:=current_totalcycles;
res:=reloj[(konami_sound.clock div 512) mod 10];
if addr(konami_sound.call)<>nil then res:=konami_sound.call(res);
konamisnd_portb:=res;
end;

function konamisnd_getbyte(direccion:word): Byte;
begin
case direccion of
  $4000:konamisnd_getbyte:=AY8910_Read(0);
  $6000:konamisnd_getbyte:=AY8910_Read(1);
    else konamisnd_getbyte:=mem_snd[direccion];
end;
end;

procedure konamisnd_putbyte(direccion:word;valor:byte);
begin
if direccion<$2000 then exit;
case direccion of
        $4000:AY8910_Write(valor,0);
        $5000:AY8910_Control(valor,0);
        $6000:AY8910_Write(valor,1);
        $7000:AY8910_Control(valor,1);
end;
mem_snd[direccion]:=valor;
end;

procedure konamisnd_2ay;
begin
  ay8910_update(0);
  ay8910_update(1);
end;

procedure konamisnd_1ay;
begin
  ay8910_update(0);
end;

procedure konamisnd_init(amp,cpu,ntipo:byte;clock:integer;call:tkonami_call);
begin
konami_sound.call:=call;
konami_sound.tipo:=ntipo;
case konami_sound.tipo of
  ONE_AY8910:begin
      init_sound_engine(cpu,clock,konamisnd_1ay);
      AY8910_init(1789772,0,amp,konamisnd_porta,konamisnd_portb,nil,nil);
    end;
  TWO_AY8910:begin
      init_sound_engine(cpu,clock,konamisnd_2ay);
      AY8910_init(1789772,0,amp,konamisnd_porta,konamisnd_portb,nil,nil);
      AY8910_init(1789772,1,amp,nil,nil,nil,nil);
    end;
end;
end;

procedure konamisnd_reset;
begin
ay8910_reset(0);
if konami_sound.tipo=TWO_AY8910 then ay8910_reset(1);
konami_sound.sound_latch:=0;
konami_sound.frame:=0;
konami_sound.clock:=0;
konami_sound.last_cycles:=0;
end;

procedure konamisnd_close;
begin
ay8910_close(0);
if konami_sound.tipo=TWO_AY8910 then ay8910_close(1);
end;

end.
