unit sn_76496;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses sound_engine;

const
 MAX_OUTPUT=$7fff;
 SN_STEP=$10000;
 FB_WNOISE=$14002;
 FB_PNOISE=$8000;
 NG_PRESET=$0f35;

type
 SN76496=record
	Clock:integer;
	UpdateStep:extended;
	VolTable:array[0..15] of integer;	// volume table
	Registers:array[0..7] of integer;	// registers
	LastRegister:integer;	// last register written
	Volume,Period,Count:array [0..3] of integer;		// volume of voice 0-2 and noise
  Output:array [0..3] of byte;
	RNG:cardinal;		// noise generator      */
	NoiseFB:integer;		// noise feedback mask */
  tsample:byte;
 end;
 psn76496=^SN76496;

var
  chip_sn:array[0..3] of psn76496;

procedure SN76496_Write(num:byte;data:integer);
procedure SN76496_init(num:byte;clock:integer);
procedure SN76496_Update(num:byte);
procedure SN76496_close(num:byte);
procedure SN76496_reset(num:byte);

implementation
uses principal;

procedure SN76496_Write(num:byte;data:integer);
var
  r,c,n:integer;
  chip:psn76496;
begin
  chip:=chip_sn[num];
	// update the output buffer before changing the registers
  //SN76496update(num);
	if (data and $80)<>0 then begin
		r:=(data and $70) shr 4;
		c:=r div 2;
		chip.LastRegister:=r;
		chip.Registers[r]:=(chip.Registers[r] and $3f0) or (data and $0f);
		case r of
			0,2,4:begin	// tone 0 : frequency,tone 1 : frequency,tone 2 : frequency */
				      chip.Period[c]:=trunc(chip.UpdateStep * chip.Registers[r]);
				      if (chip.Period[c]=0) then chip.Period[c]:=trunc(chip.UpdateStep);
				      if (r=4) then begin //update noise shift frequency */
					      if ((chip.Registers[6] and $03)=$03) then chip.Period[3]:=chip.Period[2]*2;
              end;
            end;
			1,3,5,7:begin	// tone 0 : volume,tone 1 : volume,tone 2 : volume,noise  : volume */
				        chip.Volume[c]:=chip.VolTable[data and $0f];
              end;
			6:begin	// noise  : frequency, mode */
					n:=chip.Registers[6];
					if (n and 4)<>0 then chip.NoiseFB:=FB_WNOISE
            else chip.NoiseFB:=FB_PNOISE;
					n:=n and 3;
					//* N/512,N/1024,N/2048,Tone #3 output */
          if ((n and 3)=3) then chip.Period[3]:=chip.Period[2]*2
            else chip.Period[3]:=trunc(chip.UpdateStep) shl (5+(n and 3));
					// reset noise shifter */
					chip.RNG:=NG_PRESET;
					chip.Output[3]:=chip.RNG and 1;
        end;
    end; //del case
	end else begin  //del if
		r:=chip.LastRegister;
    c:=r div 2;
		case r of
			0,2,4:begin	// tone 0 : frequency,tone 1 : frequency,tone 2 : frequency */
				      chip.Registers[r]:= (chip.Registers[r] and $0f) or ((data and $3f) shl 4);
				      chip.Period[c]:=trunc(chip.UpdateStep*chip.Registers[r]);
				      if (chip.Period[c]=0) then chip.Period[c]:=trunc(chip.UpdateStep);
				      if (r=4) then begin // update noise shift frequency */
					        if ((chip.Registers[6] and $03) = $03) then chip.Period[3]:=chip.Period[2]*2;
				      end;
            end;
      1,3,5,7:begin	//* tone 0 : volume,tone 1 : volume,tone 2 : volume,noise  : volume */
				        chip.volume[c]:= chip.VolTable[data and $0f];
				        chip.Registers[r]:=(chip.Registers[r] and $3f0) or (data and $0f);
              end;
			6:begin	// noise  : frequency, mode */
					chip.Registers[r]:= (chip.Registers[r] and $3f0) or (data and $0f);
					n:= chip.Registers[6];
					if (n and 4)<>0 then chip.NoiseFB:=FB_WNOISE
            else chip.NoiseFB:=FB_PNOISE;
					n:=n and 3;
					// N/512,N/1024,N/2048,Tone #3 output */
					if ((n and 3)=3) then chip.Period[3]:=2 * chip.Period[2]
            else chip.Period[3]:=trunc(chip.UpdateStep) shl (5+(n and 3));
					// reset noise shifter */
					chip.RNG:= NG_PRESET;
					chip.Output[3]:= chip.RNG and 1;
				end;
    end;  //del case
	end;
end;

procedure SN76496_set_gain(num:byte;gain:integer);
var
  i:integer;
  out_sn:double;
  chip:psn76496;
begin
  chip:=chip_sn[num];
	// increase max output basing on gain (0.2 dB per step) */
	out_sn:=MAX_OUTPUT / 3;
	while (gain> 0) do begin
    gain:=gain-1;
		out_sn:=out_sn*1.023292992;	// = (10 ^ (0.2/20)) */
  end;
  // build volume table (2dB per step) */
  for i:=0 to 14 do begin
    // limit volume to avoid clipping */
    if (out_sn > (MAX_OUTPUT/3)) then chip.VolTable[i]:=round(MAX_OUTPUT/3)
      else chip.VolTable[i]:=round(out_sn);
    out_sn:=out_sn/1.258925412;	// = 10 ^ (2/20) = 2dB */
  end;
  chip.VolTable[15]:=0;
end;

procedure SN76496_close(num:byte);
begin
if chip_sn[num]<>nil then begin
  freemem(chip_sn[num]);
  chip_sn[num]:=nil;
end;
end;

procedure SN76496_reset(num:byte);
var
  chip:psn76496;
  i:byte;
begin
chip:=chip_sn[num];
for i:=0 to 3 do chip.Volume[i]:=0;
chip.LastRegister:= 0;
for i:=0 to 3 do begin
  chip.Registers[i*2]:=0;
  chip.Registers[(i*2) + 1]:=$0f;	// volume = 0 */
end;
for i:=0 to 3 do begin
  chip.Output[i]:= 0;
  chip.Period[i]:=trunc(chip.UpdateStep);
  chip.Count[i]:=trunc(chip.UpdateStep);
end;
chip.RNG:=NG_PRESET;
chip.Output[3]:= chip.RNG and 1;
end;

procedure SN76496_resample(num:byte);
var
  tmp:int64;
  chip:psn76496;
begin
   chip:=chip_sn[num];
	{ the base clock for the tone generators is the chip clock divided by 16
	for the noise generator, it is clock / 256.
	Here we calculate the number of steps which happen during one sample
	at the given sample rate. No. of events = sample rate / (clock/16).
	STEP is a multiplier used to turn the fraction into a fixed point
	number. }
  tmp:=SN_STEP*16;
  tmp:=tmp*44100;
  tmp:=round(tmp / chip.clock);
	chip.UpdateStep:=tmp;
end;

procedure SN76496_init(num:byte;clock:integer);
var
  chip:psn76496;
begin
  if chip_sn[num]=nil then begin
    getmem(chip_sn[num],sizeof(sn76496));
    fillchar(chip_sn[num]^,sizeof(sn76496),0);
  end;
  chip:=chip_sn[num];
  SN76496_set_gain(num,0);
	chip.clock:=clock;
  chip.tsample:=init_channel;
	SN76496_resample(num);
  SN76496_reset(num);
end;

procedure SN76496_Update(num:byte);
Var
  i,left,nextevent:Integer;
  out_sn:double;
  vol:array[0..3] of integer;
  chip:pSN76496;
begin
   chip:=chip_sn[num];
	// If the volume is 0, increase the counter */
	for i:=0 to 3 do begin
		if (chip.Volume[i]=0) then begin
			{ note that I do count += length, NOT count = length + 1. You might think */
			 it's the same since the volume is 0, but doing the latter could cause */
			 interferencies when the program is rapidly modulating the volume. }
         if (chip.Count[i]<=SN_STEP) then inc(chip.Count[i],SN_STEP);
		end;
   end;
		{ vol[] keeps track of how long each square wave stays
		 in the 1 position during the sample period. }
    for i:=0 to 3 do vol[i]:=0;
		for i:=0 to 2 do begin
			if (chip.Output[i])<>0 then inc(vol[i],chip.Count[i]);
			dec(chip.Count[i],SN_STEP);
			{ Period[i] is the half period of the square wave. Here, in each */
			loop I add Period[i] twice, so that at the end of the loop the */
			square wave is in the same status (0 or 1) it was at the start. */
			vol[i] is also incremented by Period[i], since the wave has been 1 */
			exactly half of the time, regardless of the initial position. */
			If we exit the loop in the middle, Output[i] has to be inverted */
			and vol[i] incremented only if the exit status of the square */
			wave is 1. }
			while (chip.Count[i] <= 0) do begin
				inc(chip.Count[i],chip.Period[i]);
				if (chip.Count[i] > 0) then begin
					chip.Output[i]:=chip.Output[i] xor 1;
					if (chip.Output[i])<>0 then inc(vol[i],chip.Period[i]);
          break;
				end;
				inc(chip.Count[i],chip.Period[i]);
				inc(vol[i],chip.Period[i]);
			end; //del while
			if (chip.Output[i])<>0 then dec(vol[i],chip.Count[i]);
		end; //del for
		left:= SN_STEP;
    repeat
			if (chip.Count[3] < left) then nextevent:=chip.Count[3]
			  else nextevent:=left;
			if (chip.Output[3])<>0 then inc(vol[3],chip.Count[3]);
			dec(chip.Count[3],nextevent);
			if (chip.Count[3] <= 0) then begin
				if (chip.RNG and 1)<>0 then chip.RNG:=chip.RNG xor chip.NoiseFB;
				chip.RNG:=chip.RNG shr 1;
				chip.Output[3]:=chip.RNG and 1;
				inc(chip.Count[3],chip.Period[3]);
				if (chip.Output[3])<>0 then inc(vol[3],chip.Period[3]);
			end;
			if (chip.Output[3])<>0 then dec(vol[3],chip.Count[3]);
			dec(left,nextevent);
    until (left=0);
		out_sn:= vol[0] * chip.Volume[0] + vol[1] * chip.Volume[1] +
				vol[2] * chip.Volume[2] + vol[3] * chip.Volume[3];
    if (out_sn>MAX_OUTPUT*SN_STEP) then out_sn:=MAX_OUTPUT*SN_STEP;
    tsample[chip.tsample,sound_status.posicion_sonido]:=trunc(out_sn/SN_STEP);
end;

end.
