unit upd7759;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     sysutils,sound_engine;

const
  FRAC_BITS=20;
  FRAC_ONE=(1 shl FRAC_BITS);
  FRAC_MASK=(FRAC_ONE-1);
  // chip states
  STATE_IDLE=0;
	STATE_DROP_DRQ=1;
	STATE_START=2;
	STATE_FIRST_REQ=3;
	STATE_LAST_SAMPLE=4;
	STATE_DUMMY1=5;
	STATE_ADDR_MSB=6;
	STATE_ADDR_LSB=7;
	STATE_DUMMY2=8;
	STATE_BLOCK_HEADER=9;
	STATE_NIBBLE_COUNT=10;
	STATE_NIBBLE_MSN=11;
	STATE_NIBBLE_LSN=12;
  upd7759_step:array[0..15,0..15] of integer=(
	( 0,  0,  1,  2,  3,   5,   7,  10,  0,   0,  -1,  -2,  -3,   -5,   -7,  -10 ),
	( 0,  1,  2,  3,  4,   6,   8,  13,  0,  -1,  -2,  -3,  -4,   -6,   -8,  -13 ),
	( 0,  1,  2,  4,  5,   7,  10,  15,  0,  -1,  -2,  -4,  -5,   -7,  -10,  -15 ),
	( 0,  1,  3,  4,  6,   9,  13,  19,  0,  -1,  -3,  -4,  -6,   -9,  -13,  -19 ),
	( 0,  2,  3,  5,  8,  11,  15,  23,  0,  -2,  -3,  -5,  -8,  -11,  -15,  -23 ),
	( 0,  2,  4,  7, 10,  14,  19,  29,  0,  -2,  -4,  -7, -10,  -14,  -19,  -29 ),
	( 0,  3,  5,  8, 12,  16,  22,  33,  0,  -3,  -5,  -8, -12,  -16,  -22,  -33 ),
	( 1,  4,  7, 10, 15,  20,  29,  43, -1,  -4,  -7, -10, -15,  -20,  -29,  -43 ),
	( 1,  4,  8, 13, 18,  25,  35,  53, -1,  -4,  -8, -13, -18,  -25,  -35,  -53 ),
	( 1,  6, 10, 16, 22,  31,  43,  64, -1,  -6, -10, -16, -22,  -31,  -43,  -64 ),
	( 2,  7, 12, 19, 27,  37,  51,  76, -2,  -7, -12, -19, -27,  -37,  -51,  -76 ),
	( 2,  9, 16, 24, 34,  46,  64,  96, -2,  -9, -16, -24, -34,  -46,  -64,  -96 ),
	( 3, 11, 19, 29, 41,  57,  79, 117, -3, -11, -19, -29, -41,  -57,  -79, -117 ),
	( 4, 13, 24, 36, 50,  69,  96, 143, -4, -13, -24, -36, -50,  -69,  -96, -143 ),
	( 4, 16, 29, 44, 62,  85, 118, 175, -4, -16, -29, -44, -62,  -85, -118, -175 ),
	( 6, 20, 36, 54, 76, 104, 144, 214, -6, -20, -36, -54, -76, -104, -144, -214 ));
  upd7759_state_table:array[0..15] of integer=( -1, -1, 0, 0, 1, 2, 2, 3, -1, -1, 0, 0, 1, 2, 2, 3);

type
  tupd7759_chip=record
	  // internal clock to output sample rate mapping */
	  pos:dword;						// current output sample position */
	  step:dword;						// step value per output sample */
	  //attotime	clock_period;				/* clock period */
	  //emu_timer *timer;						/* timer */
	  // I/O lines */
	  fifo_in:byte;					// last data written to the sound chip */
	  reset:byte;						// current state of the RESET line */
	  start:byte;						// current state of the START line */
	  drq:byte;						// current state of the DRQ line */
	//void (*drqcallback)(device_t *device, int param);			/* drq callback */
	// internal state machine */
	  state:byte;						// current overall chip state */
  	clocks_left:integer;				// number of clocks left in this state */
    nibbles_left:word;				// number of ADPCM nibbles left to process */
    repeat_count:byte;				// number of repeats remaining in current repeat block */
    post_drq_state:shortint;				// state we will be in after the DRQ line is dropped */
    post_drq_clocks:integer;			// clocks that will be left after the DRQ line is dropped */
  	req_sample:byte;					// requested sample number */
    last_sample:byte;				// last sample number available */
    block_header:byte;				// header byte *
    sample_rate:byte;				// number of UPD clocks per ADPCM nibble */
    first_valid_header:byte;			// did we get our first valid header yet? */
    offset:dword;						// current ROM offset */
    repeat_offset:dword;				// current ROM repeat offset */
  	// ADPCM processing */
    adpcm_state:shortint;				// ADPCM state index */
  	adpcm_data:byte;					// current byte of ADPCM data */
    sample:integer;						// current sample value */
  	// ROM access */
  	rom:pbyte;						// pointer to ROM data or NULL for slave mode */
  	rombase:pbyte;					// pointer to ROM data or NULL for slave mode */
  	romoffset:pdword;					// ROM offset to make save/restore easier */
    resample_pos:double;
    resample_inc:double;
    tsample:byte;
    amp:byte;
  end;
  ptupd7759_chip=^tupd7759_chip;

var
  upd7759_chip:array[0..1] of ptupd7759_chip;

procedure upd7759_Init(num,amp:byte;clock:integer);
procedure upd7759_reset(num:byte);
procedure upd7759_close(num:byte);
procedure upd7759_reset_w(num,data:byte);
procedure upd7759_start_w(num,data:byte);
procedure upd7759_port_w(num,data:byte);
function upd7759_busy_r(num:byte):byte;
procedure upd7759_update(num:byte);

implementation

procedure upd7759_Init(num,amp:byte;clock:integer);
begin
  getmem(upd7759_chip[num],sizeof(tupd7759_chip));
	// compute the stepping rate based on the chip's clock speed */
	upd7759_chip[num].step:=4*FRAC_ONE;
	// compute the clock period */
	//chip->clock_period = ATTOTIME_IN_HZ(device->clock());
	// compute the ROM base or allocate a timer */
	getmem(upd7759_chip[num].rom,$20000);
	// set the DRQ callback */
	//chip->drqcallback = intf->drqcallback;
	// assume /RESET and /START are both high */
	upd7759_chip[num].reset:=1;
	upd7759_chip[num].start:=1;
  upd7759_chip[num].resample_inc:=clock/4/44100;
	// toggle the reset line to finish the reset */
  upd7759_chip[num].tsample:=init_channel;
  upd7759_chip[num].amp:=amp;
	upd7759_reset(num);
end;

procedure upd7759_close(num:byte);
begin
if upd7759_chip[num]<>nil then begin
  freemem(upd7759_chip[num].rom);
  upd7759_chip[num].rom:=nil;
  freemem(upd7759_chip[num]);
  upd7759_chip[num]:=nil;
end;
end;

procedure update_adpcm(num:byte;data:integer);inline;
var
  chip:ptupd7759_chip;
begin
	// update the sample and the state */
  chip:=upd7759_chip[num];
	chip.sample:=chip.sample+upd7759_step[chip.adpcm_state,data];
	chip.adpcm_state:=chip.adpcm_state+upd7759_state_table[data];
	// clamp the state to 0..15 */
	if (chip.adpcm_state<0) then chip.adpcm_state:=0
	  else if (chip.adpcm_state>15) then chip.adpcm_state:=15;
end;

procedure advance_state(num:byte);inline;
var
  chip:ptupd7759_chip;
begin
  chip:=upd7759_chip[num];
	case chip.state of
		// Idle state: we stick around here while there's nothing to do */
		STATE_IDLE:chip.clocks_left:=4;
		// drop DRQ state: update to the intended state */
		STATE_DROP_DRQ:begin
			  chip.drq:=0;
			  chip.clocks_left:=chip.post_drq_clocks;
			  chip.state:=chip.post_drq_state;
      end;
		// Start state: we begin here as soon as a sample is triggered */
		STATE_START:begin
        chip.resample_pos:=0;
			  if chip.rom<>nil then chip.req_sample:=chip.fifo_in
          else chip.req_sample:=$10;
		  	{ 35+ cycles after we get here, the /DRQ goes low
             *     (first byte (number of samples in ROM) should be sent in response)
             *
             * (35 is the minimum number of cycles I found during heavy tests.
             * Depending on the state the chip was in just before the /MD was set to 0 (reset, standby
             * or just-finished-playing-previous-sample) this number can range from 35 up to ~24000).
             * It also varies slightly from test to test, but not much - a few cycles at most.) }
	  		chip.clocks_left:=70;	// 35 - breaks cotton */
  			chip.state:=STATE_FIRST_REQ;
      end;
		// First request state: issue a request for the first byte */
		// The expected response will be the index of the last sample */
		STATE_FIRST_REQ:begin
			  chip.drq:=1;
  			// 44 cycles later, we will latch this value and request another byte */
  			chip.clocks_left:=44;
  			chip.state:=STATE_LAST_SAMPLE;
			end;
		// Last sample state: latch the last sample value and issue a request for the second byte */
		// The second byte read will be just a dummy */
		STATE_LAST_SAMPLE:begin
  			if chip.rom<>nil then chip.last_sample:=pbytearray(chip.rom)^[0]
          else chip.last_sample:=chip.fifo_in;
  			chip.drq:= 1;
  			// 28 cycles later, we will latch this value and request another byte */
  			chip.clocks_left:=28;	// 28 - breaks cotton */
  			if (chip.req_sample>chip.last_sample) then chip.state:=STATE_IDLE
          else chip.state:=STATE_DUMMY1;
			end;
		// First dummy state: ignore any data here and issue a request for the third byte */
		// The expected response will be the MSB of the sample address */
		STATE_DUMMY1:begin
  			chip.drq:=1;
  			// 32 cycles later, we will latch this value and request another byte */
	  		chip.clocks_left:= 32;
		  	chip.state:=STATE_ADDR_MSB;
			end;
		// Address MSB state: latch the MSB of the sample address and issue a request for the fourth byte */
		// The expected response will be the LSB of the sample address */
		STATE_ADDR_MSB:begin
			if chip.rom<>nil then chip.offset:=(pbytearray(chip.rom)^[chip.req_sample*2+5]) shl 9
        else chip.offset:=chip.fifo_in shl 9;
			chip.drq:=1;
			// 44 cycles later, we will latch this value and request another byte */
			chip.clocks_left:=44;
			chip.state:=STATE_ADDR_LSB;
    end;
		// Address LSB state: latch the LSB of the sample address and issue a request for the fifth byte */
		// The expected response will be just a dummy */
		STATE_ADDR_LSB:begin
			if chip.rom<>nil then chip.offset:=chip.offset or ((pbytearray(chip.rom)^[chip.req_sample*2+6]) shl 1)
        else chip.offset:=chip.offset or (chip.fifo_in shl 1);
			chip.drq:=1;
			// 36 cycles later, we will latch this value and request another byte */
			chip.clocks_left:=36;
			chip.state:=STATE_DUMMY2;
    end;
		// Second dummy state: ignore any data here and issue a request for the the sixth byte */
		// The expected response will be the first block header */
		STATE_DUMMY2:begin
			chip.offset:=chip.offset+1;
			chip.first_valid_header:=0;
			chip.drq:=1;
			// 36?? cycles later, we will latch this value and request another byte */
			chip.clocks_left:=36;
			chip.state:=STATE_BLOCK_HEADER;
    end;
		// Block header state: latch the header and issue a request for the first byte afterwards */
		STATE_BLOCK_HEADER:begin
			// if we're in a repeat loop, reset the offset to the repeat point and decrement the count */
			if (chip.repeat_count<>0) then begin
				chip.repeat_count:=chip.repeat_count-1;
				chip.offset:=chip.repeat_offset;
			end;
			if chip.rom<>nil then chip.block_header:=pbytearray(chip.rom)^[chip.offset and $1ffff]
          else chip.block_header:=chip.fifo_in;
      chip.offset:=chip.offset+1;
			chip.drq:=1;
			// our next step depends on the top two bits */
			case (chip.block_header and $c0) of
				$00:begin	// silence */
    					chip.clocks_left:=1024*((chip.block_header and $3f)+1);
              if ((chip.block_header=0) and ((chip.first_valid_header and 1)<>0)) then chip.state:=STATE_IDLE
                else chip.state:=STATE_BLOCK_HEADER;
    					chip.sample:=0;
    					chip.adpcm_state:=0;
            end;
				$40:begin	// 256 nibbles */
    					chip.sample_rate:=(chip.block_header and $3f)+1;
    					chip.nibbles_left:= 256;
    					chip.clocks_left:=36;	// just a guess */
    					chip.state:=STATE_NIBBLE_MSN;
            end;
	  		$80:begin	// n nibbles */
    					chip.sample_rate:=(chip.block_header and $3f)+1;
    					chip.clocks_left:=36;	// just a guess */
    					chip.state:=STATE_NIBBLE_COUNT;
  					end;
				$c0:begin	// repeat loop */
    					chip.repeat_count:=(chip.block_header and 7)+1;
    					chip.repeat_offset:=chip.offset;
    					chip.clocks_left:=36;	// just a guess */
    					chip.state:=STATE_BLOCK_HEADER;
  					end;
			end;
			// set a flag when we get the first non-zero header */
			if (chip.block_header<>0) then chip.first_valid_header:=1;
    end;

		// Nibble count state: latch the number of nibbles to play and request another byte */
		// The expected response will be the first data byte */
		STATE_NIBBLE_COUNT:begin
    			if (chip.rom<>nil) then chip.nibbles_left:=(pbytearray(chip.rom)^[chip.offset and $1ffff])+1
            else chip.nibbles_left:=chip.fifo_in+1;
          chip.offset:=chip.offset+1;
    			chip.drq:= 1;
    			// 36?? cycles later, we will latch this value and request another byte */
    			chip.clocks_left:=36;	// just a guess */
    			chip.state:=STATE_NIBBLE_MSN;
			end;
		// MSN state: latch the data for this pair of samples and request another byte */
		// The expected response will be the next sample data or another header */
		STATE_NIBBLE_MSN:begin
			if chip.rom<>nil then chip.adpcm_data:=pbytearray(chip.rom)^[chip.offset and $1ffff]
        else chip.adpcm_data:=chip.fifo_in;
      chip.offset:=chip.offset+1;
			update_adpcm(num,chip.adpcm_data shr 4);
			chip.drq:=1;
			// we stay in this state until the time for this sample is complete */
			chip.clocks_left:=chip.sample_rate*4;
      chip.nibbles_left:=chip.nibbles_left-1;
			if chip.nibbles_left=0 then chip.state:=STATE_BLOCK_HEADER
  			else chip.state:=STATE_NIBBLE_LSN;
    end;
		// LSN state: process the lower nibble */
		STATE_NIBBLE_LSN:begin
			update_adpcm(num,chip.adpcm_data and 15);
			// we stay in this state until the time for this sample is complete */
			chip.clocks_left:=chip.sample_rate*4;
      chip.nibbles_left:=chip.nibbles_left-1;
			if (chip.nibbles_left=0) then chip.state:=STATE_BLOCK_HEADER
  			else chip.state:=STATE_NIBBLE_MSN;
    end;
	end;
	// if there's a DRQ, fudge the state */
	if ((chip.drq and 1)<>0) then begin
		chip.post_drq_state:=chip.state;
		chip.post_drq_clocks:=chip.clocks_left-21;
		chip.state:=STATE_DROP_DRQ;
		chip.clocks_left:=21;
	end;
end;

procedure upd7759_reset(num:byte);
var
  chip:ptupd7759_chip;
begin
  chip:=upd7759_chip[num];
	chip.pos               := 0;
	chip.fifo_in           := 0;
	chip.drq               := 0;
	chip.state             := STATE_IDLE;
	chip.clocks_left       := 0;
	chip.nibbles_left      := 0;
	chip.repeat_count      := 0;
	chip.post_drq_state    := STATE_IDLE;
	chip.post_drq_clocks   := 0;
	chip.req_sample        := 0;
	chip.last_sample       := 0;
	chip.block_header      := 0;
	chip.sample_rate       := 0;
	chip.first_valid_header:= 0;
	chip.offset            := 0;
	chip.repeat_offset     := 0;
	chip.adpcm_state       := 0;
	chip.adpcm_data        := 0;
	chip.sample            := 0;

{	/* turn off any timer */
	if (chip.timer)
		timer_adjust_oneshot(chip.timer, attotime_never, 0);}
end;

procedure upd7759_reset_w(num,data:byte);
var
  chip:ptupd7759_chip;
  oldreset:byte;
begin
  chip:=upd7759_chip[num];
	// update the reset value */
	oldreset:=chip.reset;
	if (data<>0) then chip.reset:=1
    else chip.reset:=0;
	// on the falling edge, reset everything */
	if ((oldreset<>0) and ((not(chip.reset) and 1)<>0)) then upd7759_reset(num);
end;

procedure upd7759_start_w(num,data:byte);
var
  chip:ptupd7759_chip;
  oldstart:byte;
begin
  chip:=upd7759_chip[num];
	// update the start value */
	oldstart:=chip.start;
	if (data<>0) then chip.start:=1
    else chip.start:=0;

	// on the rising edge, if we're idle, start going, but not if we're held in reset */
	if ((chip.state=STATE_IDLE) and ((not(oldstart) and 1)<>0) and ((chip.start and 1)<>0) and ((chip.reset and 1)<>0)) then begin
		chip.state:=STATE_START;
		// for slave mode, start the timer going */
 {		if (chip->timer)
			timer_adjust_oneshot(chip->timer, attotime_zero, 0);}
	end;
end;

procedure upd7759_port_w(num,data:byte);
begin
	// update the FIFO value */
	upd7759_chip[num].fifo_in:=data;
end;

function upd7759_busy_r(num:byte):byte;
begin
	// return /BUSY */
  if upd7759_chip[num].state=STATE_IDLE then upd7759_busy_r:=1
    else upd7759_busy_r:=0;
end;

procedure upd7759_update(num:byte);
var
  chip:ptupd7759_chip;
  clocks_this_time,clocks_left:integer;
  step,pos:dword;
  out_:integer;
  num_samples:byte;
begin
  chip:=upd7759_chip[num];
	clocks_left:=chip.clocks_left;
  step:=chip.step;
  pos:=chip.pos;
  out_:=0;
  chip.resample_pos:=chip.resample_pos+chip.resample_inc;
  num_samples:=trunc(chip.resample_pos);
  chip.resample_pos:=chip.resample_pos-num_samples;
	// loop until done */
	if (chip.state<>STATE_IDLE) then begin
    while (num_samples<>0) do begin
			// store the current sample */
			out_:=chip.sample shl 7;
			// advance by the number of clocks/output sample */
			pos:=pos+step;
			// handle clocks, but only in standalone mode */
			while ((chip.rom<>nil) and (pos>=FRAC_ONE)) do begin
				clocks_this_time:=pos shr FRAC_BITS;
				if (clocks_this_time>clocks_left) then clocks_this_time:=clocks_left;
				// clock once */
				pos:=pos-(clocks_this_time*FRAC_ONE);
				clocks_left:=clocks_left-clocks_this_time;
				// if we're out of clocks, time to handle the next state */
				if (clocks_left=0) then begin
					// advance one state; if we hit idle, bail */
					advance_state(num);
					if (chip.state=STATE_IDLE) then break;
					// reimport the variables that we cached */
					clocks_left:=chip.clocks_left;
          out_:=(out_+(chip.sample*64)) div 2;
				end;
      end;
      num_samples:=num_samples-1;
	  end;
  end;
	// flush the state back */
	chip.clocks_left:= clocks_left;
	chip.pos:=pos;
  if out_>32767 then out_:=32767
    else if out_<-32768 then out_:=-32768;
  tsample[chip.tsample,sound_status.posicion_sonido]:=out_*chip.amp;
  if sound_status.stereo then tsample[chip.tsample,sound_status.posicion_sonido+1]:=out_*chip.amp;
end;

end.
