unit vlm_5030;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses {$IFDEF WINDOWS}windows,{$ENDIF}
     sysutils,sound_engine,timer_engine;

const
  FR_SIZE=4;
  // samples per interpolator */
  IP_SIZE_SLOWER=(240 div FR_SIZE);
  IP_SIZE_SLOW=(200 div FR_SIZE);
  IP_SIZE_NORMAL=(160 div FR_SIZE);
  IP_SIZE_FAST=(120 div FR_SIZE);
  IP_SIZE_FASTER=(80 div FR_SIZE);

	PH_RESET=0;
	PH_IDLE=1;
	PH_SETUP=2;
	PH_WAIT=3;
	PH_RUN=4;
	PH_STOP=5;
	PH_END=6;

// ROM Tables */
VLM5030_speed_table:array[0..8-1] of integer=(
 IP_SIZE_NORMAL,
 IP_SIZE_FAST,
 IP_SIZE_FASTER,
 IP_SIZE_FASTER,
 IP_SIZE_NORMAL,
 IP_SIZE_SLOWER,
 IP_SIZE_SLOW,
 IP_SIZE_SLOW
);

// This is the energy lookup table */
// sampled from real chip */
  energytable:array[0..$20-1] of byte=(
	  0,  2,  4,  6, 10, 12, 14, 18, // 0-7
	 22, 26, 30, 34, 38, 44, 48, 54, //  8-15
	 62, 68, 76, 84, 94,102,114,124, // 16-23
	136,150,164,178,196,214,232,254  // 24-31
  );
// This is the pitch lookup table */
  pitchtable:array[0..$20-1] of byte=(
   1,                               // 0     : random mode */
   22,                              // 1     : start=22    */
   23, 24, 25, 26, 27, 28, 29, 30,  //  2- 9 : 1step       */
   32, 34, 36, 38, 40, 42, 44, 46,  // 10-17 : 2step       */
   50, 54, 58, 62, 66, 70, 74, 78,  // 18-25 : 4step       */
   86, 94, 102,110,118,126          // 26-31 : 8step       */
   );

  K1_table:array[0..63] of integer= (
  -24898,  -25672,  -26446,  -27091,  -27736,  -28252,  -28768,  -29155,
  -29542,  -29929,  -30316,  -30574,  -30832,  -30961,  -31219,  -31348,
  -31606,  -31735,  -31864,  -31864,  -31993,  -32122,  -32122,  -32251,
  -32251,  -32380,  -32380,  -32380,  -32509,  -32509,  -32509,  -32509,
   24898,   23995,   22963,   21931,   20770,   19480,   18061,   16642,
   15093,   13416,   11610,    9804,    7998,    6063,    3999,    1935,
       0,   -1935,   -3999,   -6063,   -7998,   -9804,  -11610,  -13416,
  -15093,  -16642,  -18061,  -19480,  -20770,  -21931,  -22963,  -23995
);

  K2_table:array[0..31] of integer= (
       0,   -3096,   -6321,   -9417,  -12513,  -15351,  -18061,  -20770,
  -23092,  -25285,  -27220,  -28897,  -30187,  -31348,  -32122,  -32638,
       0,   32638,   32122,   31348,   30187,   28897,   27220,   25285,
   23092,   20770,   18061,   15351,   12513,    9417,    6321,    3096
);
  K3_table:array[0..15] of integer= (
       0,   -3999,   -8127,  -12255,  -16384,  -20383,  -24511,  -28639,
   32638,   28639,   24511,   20383,   16254,   12255,    8127,    3999
);
  K5_table:array[0..7] of integer= (
       0,   -8127,  -16384,  -24511,   32638,   24511,   16254,    8127
);


type
  vlm5030_info=record
	  rom:pbyte;
	  address_mask:integer;
	  address:word;
	  pin_BSY:byte;
	  pin_ST:byte;
	  pin_VCU:byte;
	  pin_RST:byte;
	  latch_data:byte;
	  vcu_addr_h:word;
	  parameter:byte;
	  phase:byte;
	  // state of option paramter */
	  frame_size:integer;
	  pitch_offset:integer;
	  interp_step:byte;

	  interp_count:byte;       // number of interp periods    */
	  sample_count:byte;       // sample number within interp */
	  pitch_count:byte;
	// these contain data describing the current and previous voice frames */
	  old_energy:word;
	  old_pitch:byte;
	  old_k:array[0..10-1] of integer;
	  target_energy:word;
	  target_pitch:byte;
	  target_k:array[0..10-1] of integer;

	  new_energy:word;
	  new_pitch:byte;
	  new_k:array[0..10-1] of integer;

	  // these are all used to contain the current state of the sound generation */
	  current_energy:cardinal;
	  current_pitch:cardinal;
	  current_k:array[0..10-1] of integer;
	  x:array[0..10-1] of integer;
    amplificador:byte;
    tsample:byte;
    out_:integer;
  end;
  pvlm5030=^vlm5030_info;
var
  chip_vlm:pvlm5030;

// get BSY pin level */
function VLM5030_BSY:integer;
// latch contoll data */
procedure VLM5030_data_w(data:byte);
// set RST pin level : reset / set table address A8-A15 */
procedure VLM5030_RST(pin:integer);
// set VCU pin level : ?? unknown */
procedure VLM5030_VCU(pin:integer);
// set ST pin level  : set table address A0-A7 / start speech */
procedure VLM5030_ST(pin:integer);
procedure vlm5030_init(amplificador:byte;clock:integer;rom_size:integer;num_cpm:byte;snd_clock:integer);
procedure vlm5030_close;
procedure VLM5030_reset;
procedure update_stream_vlm5030;
procedure vlm5030_update;

implementation

function get_bits(sbit,bits:integer):integer;inline;
var
  data,offset:integer;
begin
	offset:= chip_vlm.address+(sbit shr 3);
	data:=pbytearray(chip_vlm.rom)^[offset and chip_vlm.address_mask];
  data:=data or (pbytearray(chip_vlm.rom)^[(offset+1) and chip_vlm.address_mask]) shl 8;
	data:=data shr (sbit and 7);
	data:=data and ($ff shr (8-bits));
	get_bits:=data;
end;

// get next frame */
function parse_frame:integer;inline;
var
  cmd:byte;
	i,nums:integer;
begin
	// remember previous frame */
	chip_vlm.old_energy:= chip_vlm.new_energy;
	chip_vlm.old_pitch:= chip_vlm.new_pitch;
	for i:=0 to 9 do chip_vlm.old_k[i]:= chip_vlm.new_k[i];
  // command byte check */
  cmd:=pbytearray(chip_vlm.rom)^[chip_vlm.address and chip_vlm.address_mask];
  if (cmd and $01)<>0 then begin
	    // extend frame */
		  chip_vlm.new_energy:=0;
      chip_vlm.new_pitch:=0;
		  for i:=0 to 9 do chip_vlm.new_k[i]:=0;
		  chip_vlm.address:=chip_vlm.address+1;
      if (cmd and $02 )<>0 then begin
			  // end of speech */
			  // logerror("VLM5030 %04X end \n",chip->address ); */
			  parse_frame:=0;
        exit;
		  end else begin
			  // silent frame */
			  nums:= ( (cmd shr 2)+1 )*2;
			  // logerror("VLM5030 %04X silent %d frame\n",chip->address,nums ); */
			  parse_frame:=nums * FR_SIZE;
        exit;
		  end;
	end;
	// pitch */
	chip_vlm.new_pitch:=(pitchtable[get_bits(1,5)] + chip_vlm.pitch_offset ) and $ff;
	// energy */
	chip_vlm.new_energy:= energytable[get_bits(6,5)];

	// 10 K's */
	chip_vlm.new_k[9]:= K5_table[get_bits(11,3)];
	chip_vlm.new_k[8]:= K5_table[get_bits(14,3)];
	chip_vlm.new_k[7]:= K5_table[get_bits(17,3)];
	chip_vlm.new_k[6]:= K5_table[get_bits(20,3)];
	chip_vlm.new_k[5]:= K5_table[get_bits(23,3)];
	chip_vlm.new_k[4]:= K5_table[get_bits(26,3)];
	chip_vlm.new_k[3]:= K3_table[get_bits(29,4)];
	chip_vlm.new_k[2]:= K3_table[get_bits(33,4)];
	chip_vlm.new_k[1]:= K2_table[get_bits(37,5)];
	chip_vlm.new_k[0]:= K1_table[get_bits(42,6)];

	chip_vlm.address:=chip_vlm.address+6;
	//logerror("VLM5030 %04X voice \n",chip->address );
	parse_frame:=FR_SIZE;
end;

// decode and buffering data */
procedure update_stream_vlm5030;
var
	interp_effect,i,current_val:integer;
	u:array[0..11-1] of integer;
  val:integer;
label
  phase_stop;
begin
	// running */
	if ((chip_vlm.phase=PH_RUN) or (chip_vlm.phase=PH_STOP)) then begin
		// playing speech */
		//while (length > 0) do begin
			// check new interpolator or  new frame */
			if (chip_vlm.sample_count=0) then begin
				if (chip_vlm.phase=PH_STOP) then begin
					chip_vlm.phase:=PH_END;
					chip_vlm.sample_count:= 1;
					goto phase_stop; // continue to end phase */
				end;
				chip_vlm.sample_count:= chip_vlm.frame_size;
				// interpolator changes */
				if (chip_vlm.interp_count=0) then begin
					// change to new frame */
					chip_vlm.interp_count:=parse_frame; // with change phase */
					if (chip_vlm.interp_count=0) then begin
					 	// end mark found */
						chip_vlm.interp_count:= FR_SIZE;
						chip_vlm.sample_count:= chip_vlm.frame_size; // end -> stop time */
						chip_vlm.phase:= PH_STOP;
					end;
					// Set old target as new start of frame */
					chip_vlm.current_energy:= chip_vlm.old_energy;
					chip_vlm.current_pitch:= chip_vlm.old_pitch;
					for i:=0 to 9 do chip_vlm.current_k[i]:=chip_vlm.old_k[i];
					// is this a zero energy frame? */
					if (chip_vlm.current_energy=0) then begin
						//mame_printf_debug("processing frame: zero energy\n");*/
						chip_vlm.target_energy:= 0;
						chip_vlm.target_pitch:= chip_vlm.current_pitch;
						for i:=0 to 9 do chip_vlm.target_k[i]:=chip_vlm.current_k[i];
					end else begin
						chip_vlm.target_energy:= chip_vlm.new_energy;
						chip_vlm.target_pitch:= chip_vlm.new_pitch;
						for i:=0 to 9 do chip_vlm.target_k[i]:=chip_vlm.new_k[i];
					end;
				end;
				// next interpolator */
				// Update values based on step values 25% , 50% , 75% , 100% */
				chip_vlm.interp_count:=chip_vlm.interp_count-chip_vlm.interp_step;
				// 3,2,1,0 -> 1,2,3,4 */
				interp_effect:= FR_SIZE - (chip_vlm.interp_count mod FR_SIZE);
				chip_vlm.current_energy:= chip_vlm.old_energy + (chip_vlm.target_energy - chip_vlm.old_energy) * interp_effect div FR_SIZE;
				if (chip_vlm.old_pitch>1) then chip_vlm.current_pitch:= chip_vlm.old_pitch + (chip_vlm.target_pitch - chip_vlm.old_pitch) * interp_effect div FR_SIZE;
				for i:=0 to 9 do chip_vlm.current_k[i]:=chip_vlm.old_k[i] + (chip_vlm.target_k[i] - chip_vlm.old_k[i]) * interp_effect div FR_SIZE;
			end;
			// calcrate digital filter */
			if (chip_vlm.old_energy=0) then begin
				// generate silent samples here */
				current_val:=0;
			end else if (chip_vlm.old_pitch <= 1) then begin
				  // generate unvoiced samples here */
          if (random(256) and 1)<>0 then current_val:=chip_vlm.current_energy
            else current_val:=-chip_vlm.current_energy;
			  end else begin
				  // generate voiced samples here */
				  if (chip_vlm.pitch_count=0) then current_val:=chip_vlm.current_energy
            else current_val:=0;
			  end;

			// Lattice filter here */
			u[10]:=current_val;
			for i:=9 downto 0 do u[i]:=u[i+1]-((chip_vlm.current_k[i] * chip_vlm.x[i]) div 32768);
			for i:=9 downto 1 do chip_vlm.x[i]:=chip_vlm.x[i-1] + ((chip_vlm.current_k[i-1] * u[i-1]) div 32768);
			chip_vlm.x[0]:=u[0];

			// clipping, buffering */
			if (u[0] > 511) then val:=511 shl 6
			  else if (u[0] < -511) then val:=-511 shl 6
			    else val:=u[0] shl 6;
      chip_vlm.out_:=val*chip_vlm.amplificador;
      if chip_vlm.out_<-32767 then chip_vlm.out_:=-32767
        else if chip_vlm.out_>32767 then chip_vlm.out_:=32767;
			// sample count */
			chip_vlm.sample_count:=chip_vlm.sample_count-1;
			// pitch */
			chip_vlm.pitch_count:=chip_vlm.pitch_count+1;
			if (chip_vlm.pitch_count >= chip_vlm.current_pitch ) then chip_vlm.pitch_count:=0;
		//end; //del while
    exit;
	end;
	// stop phase */
phase_stop:
	case (chip_vlm.phase) of
	  PH_SETUP:if (chip_vlm.sample_count<=1) then begin
			        chip_vlm.sample_count:=0;
			        // logerror("VLM5030 BSY=H\n" ); */
			        // pin_BSY = 1; */
			        chip_vlm.phase:=PH_WAIT;
		         end else begin
			        chip_vlm.sample_count:=chip_vlm.sample_count-1;
		         end;
	  PH_END:if (chip_vlm.sample_count<=1) then begin
			      chip_vlm.sample_count:= 0;
			      // logerror("VLM5030 BSY=L\n" ); */
			      chip_vlm.pin_BSY:= 0;
			      chip_vlm.phase:= PH_IDLE;
		      end	else begin
			      chip_vlm.sample_count:=chip_vlm.sample_count-1;
		      end;
	end;
	// silent buffering
  chip_vlm.out_:=0;
end;

// get BSY pin level */
function VLM5030_BSY:integer;
begin
	VLM5030_BSY:=chip_vlm.pin_BSY;
end;

// latch contoll data */
procedure VLM5030_data_w(data:byte);
begin
	chip_vlm.latch_data:=data;
end;

// setup parameteroption when RST=H */
procedure VLM5030_setup_parameter(param:byte);inline;
begin
	// latch parameter value */
	chip_vlm.parameter:=param;

	// bit 0,1 : 4800bps / 9600bps , interporator step */
	if (param and 2)<>0 then // bit 1 = 1 , 9600bps */
		chip_vlm.interp_step:= 4 // 9600bps : no interporator */
	else if(param and 1)<>0 then // bit1 = 0 & bit0 = 1 , 4800bps */
		chip_vlm.interp_step:= 2 // 4800bps : 2 interporator */
	else	// bit1 = bit0 = 0 : 2400bps */
		chip_vlm.interp_step:= 1; // 2400bps : 4 interporator */

	// bit 3,4,5 : speed (frame size) */
	chip_vlm.frame_size:=VLM5030_speed_table[(param shr 3) and 7];

	// bit 6,7 : low / high pitch */
	if (param and $80)<>0 then	// bit7=1 , high pitch */
		chip_vlm.pitch_offset:= -8
	else if (param and $40)<>0 then	// bit6=1 , low pitch */
		chip_vlm.pitch_offset:= 8
	else
		chip_vlm.pitch_offset:= 0;
end;

procedure VLM5030_reset;
begin
	chip_vlm.phase:= PH_RESET;
	chip_vlm.address:= 0;
	chip_vlm.vcu_addr_h:= 0;
	chip_vlm.pin_BSY:= 0;

	chip_vlm.old_energy:=0;
  chip_vlm.old_pitch:=0;
	chip_vlm.new_energy:=0;
  chip_vlm.new_pitch:=0;
	chip_vlm.current_energy:=0;
  chip_vlm.current_pitch:= 0;
	chip_vlm.target_energy:=0;
  chip_vlm.target_pitch:= 0;

	fillchar(chip_vlm.old_k[0],10*4,0);
	fillchar(chip_vlm.new_k[0],10*4,0);
	fillchar(chip_vlm.current_k,10*4,0);
	fillchar(chip_vlm.target_k,10*4,0);
	chip_vlm.interp_count:=0;
  chip_vlm.sample_count:=0;
  chip_vlm.pitch_count:= 0;
	fillchar(chip_vlm.x[0],10*4,0);
	// reset parameters */
	VLM5030_setup_parameter($00);
end;


// set RST pin level : reset / set table address A8-A15 */
procedure VLM5030_RST(pin:integer);
begin
	if chip_vlm.pin_RST<>0 then begin
		if (pin=0) then begin
			// H -> L : latch parameters */
			chip_vlm.pin_RST:= 0;
			VLM5030_setup_parameter(chip_vlm.latch_data);
		end;
	end else begin
		if (pin<>0) then begin
			// L -> H : reset chip */
			chip_vlm.pin_RST:= 1;
			if chip_vlm.pin_BSY<>0 then begin
				VLM5030_reset;
			end;
		end;
	end;
end;

// set VCU pin level : ?? unknown */
procedure VLM5030_VCU(pin:integer);
begin
	// direct mode / indirect mode */
	chip_vlm.pin_VCU:= pin;
end;

// set ST pin level  : set table address A0-A7 / start speech */
procedure VLM5030_ST(pin:integer);
var
  table:integer;
begin
	if (chip_vlm.pin_ST<>pin) then begin
		// pin level is change */
		if (pin=0) then begin
			// H -> L */
			chip_vlm.pin_ST:= 0;
			if( chip_vlm.pin_VCU<>0 ) then begin
				// direct access mode & address High */
				chip_vlm.vcu_addr_h:=(chip_vlm.latch_data shl 8) + $01;
			end	else begin
				// start speech */
				// check access mode */
				if (chip_vlm.vcu_addr_h<>0) then begin
					// direct access mode */
					chip_vlm.address:= (chip_vlm.vcu_addr_h and $ff00) + chip_vlm.latch_data;
					chip_vlm.vcu_addr_h:= 0;
				end	else begin
					// indirect accedd mode */
					table:= (chip_vlm.latch_data and $fe) + ((chip_vlm.latch_data and 1) shl 8);
          chip_vlm.address:=(pbytearray(chip_vlm.rom)^[table and chip_vlm.address_mask]) shl 8;
          chip_vlm.address:=chip_vlm.address+(pbytearray(chip_vlm.rom)^[(table+1) and chip_vlm.address_mask]);
				end;
				// logerror("VLM5030 %02X start adr=%04X\n",table/2,chip->address ); */
				// reset process status */
				chip_vlm.sample_count:= chip_vlm.frame_size;
				chip_vlm.interp_count:= FR_SIZE;
				// clear filter */
				// start after 3 sampling cycle */
				chip_vlm.phase:= PH_RUN;
			end;
		end	else begin
			// L -> H */
			chip_vlm.pin_ST:= 1;
			// setup speech , BSY on after 30ms? */
			chip_vlm.phase:= PH_SETUP;
			chip_vlm.sample_count:= 1; // wait time for busy on */
			chip_vlm.pin_BSY:= 1;
		end;
	end;
end;

// start VLM5030 with sound rom              */
procedure vlm5030_init(amplificador:byte;clock:integer;rom_size:integer;num_cpm:byte;snd_clock:integer);
begin
  if chip_vlm=nil then begin
    getmem(chip_vlm,sizeof(vlm5030_info));
    fillchar(chip_vlm^,sizeof(vlm5030_info),0);
    getmem(chip_vlm.rom,rom_size);
  end;
	//emulation_rate:= clock / 440;
	// reset input pins */
	chip_vlm.pin_RST:=0;
  chip_vlm.pin_ST:=0;
  chip_vlm.pin_VCU:= 0;
	chip_vlm.latch_data:= 0;
	VLM5030_reset;
	chip_vlm.phase:= PH_IDLE;
  chip_vlm.amplificador:=amplificador;
	// memory size */
	chip_vlm.address_mask:=rom_size-1;
  chip_vlm.tsample:=init_channel;
  //timer interno
  init_timer(num_cpm,snd_clock/(clock/440),update_stream_vlm5030,true);
  init_timer(num_cpm,snd_clock/44100,vlm5030_update,true);
  chip_vlm.out_:=0;
end;

procedure vlm5030_update;
begin
tsample[chip_vlm.tsample,sound_status.posicion_sonido]:=chip_vlm.out_;
end;

procedure vlm5030_close;
begin
if chip_vlm<>nil then begin
    freemem(chip_vlm.rom);
    chip_vlm.rom:=nil;
    freemem(chip_vlm);
    chip_vlm:=nil;
end;
end;

end.
