unit ym_2203;

{$ifdef fpc}{$mode delphi}{$H+}{$endif}

interface
uses fmopn,ay_8910,timer_engine,sound_engine;

const
  MAX_YM2203=3;
type
  tipo_YM2203=record
     OPN:pfm_opn;
     REGS:array[0..255] of byte;
     tsample:byte;
     timer_adjust:double;
     timer1,timer2:byte;
  end;
  ptipo_YM2203=^tipo_YM2203;
  IRQ_Handler=procedure (numchip:integer;irqstate:integer);

var
  FM2203:array[0..MAX_YM2203] of ptipo_YM2203;

procedure ym2203_init(num,num_cpu:byte;clock,snd_clock:integer;ay_amp:byte;porta_read,portb_read:read_port;porta_write,portb_write:write_port;irq_func:IRQ_Handler);
procedure ym2203_close(num:byte);
function ym2203_read_status(num:byte):byte;
function ym2203_read_reg(num:byte):byte;
procedure ym2203_control(num,data:byte);
procedure ym2203_write_reg(num,data:byte);
procedure YM2203_Update(num:byte);
procedure YM2203_Reset(num:byte);
procedure ym2203_timer1_0;
procedure ym2203_timer2_0;
procedure ym2203_timer1_1;
procedure ym2203_timer2_1;
procedure YM2203_init_timer_a(num:byte;count:integer);
procedure YM2203_init_timer_b(num:byte;count:integer);

implementation
uses principal;

procedure reset_channels(num,chan:byte);
var
  c,s:byte;
  OPN:pfm_opn;
begin
  opn:=FM2203[num].opn;
	for c:=0 to chan-1 do begin
		opn.p_CH[c].fc:=0;
		for s:=0 to 3 do begin
			opn.p_CH[c].SLOT[s].ssg:=0;
			opn.p_CH[c].SLOT[s].ssgn:=0;
			opn.p_CH[c].SLOT[s].state:=EG_OFF;
			opn.p_CH[c].SLOT[s].volume:=MAX_ATT_INDEX;
			opn.p_CH[c].SLOT[s].vol_out:=MAX_ATT_INDEX;
		end;
	end;
end;

procedure YM2203_Reset(num:byte);
var
  i:byte;
  OPN:pfm_opn;
begin
		OPN:=FM2203[num].OPN;
		// Reset Prescaler
	  OPNPrescaler_w(OPN,0,1);
	  // reset SSG section */
	  AY8910_reset(num);
	  // status clear */
	  FM_IRQMASK_SET(OPN.ST,$03);
	  OPNWriteMode(num,OPN,$27,$30); //mode 0 , timer reset
	  OPN.eg_timer:=0;
	  OPN.eg_cnt:=0;
	  FM_STATUS_RESET(OPN.ST,$ff);
    opn.st.mode:=0;	//normal mode
  	opn.st.TA:=0;
  	opn.ST.TAC:=0;
  	opn.ST.TB:=0;
  	opn.ST.TBC:=0;
	  reset_channels(num,4);
		// reset OPerator paramater */
		for i:=$b2 downto $30 do OPNWriteReg(OPN,i,0);
		for i:=$26 downto $20 do OPNWriteReg(OPN,i,0);
end;

procedure ym2203_close(num:byte);
begin
if FM2203[num]<>nil then begin
  AY8910_close(num);
  //Cierro el OPN
  opn_close(FM2203[num].OPN);
  FM2203[num].OPN:=nil;
  freemem(FM2203[num]);
  FM2203[num]:=nil;
end;
end;

procedure ym2203_init_timers(sound_clock:integer;num_snd_cpu:byte;num:byte);
begin
  //Timers
  FM2203[num].timer_adjust:=sound_clock/FM2203[num].OPN.ST.clock;
  case num of
    0:begin
        FM2203[num].timer1:=init_timer(num_snd_cpu,1,ym2203_timer1_0,false);
        FM2203[num].timer2:=init_timer(num_snd_cpu,1,ym2203_timer2_0,false);
      end;
    1:begin
        FM2203[num].timer1:=init_timer(num_snd_cpu,1,ym2203_timer1_1,false);
        FM2203[num].timer2:=init_timer(num_snd_cpu,1,ym2203_timer2_1,false);
      end;
  end;
end;

procedure ym2203_init(num,num_cpu:byte;clock,snd_clock:integer;ay_amp:byte;porta_read,portb_read:read_port;porta_write,portb_write:write_port;irq_func:IRQ_Handler);
begin
  getmem(FM2203[num],sizeof(tipo_ym2203));
  //El PSG
  AY8910_init_internal(clock,num,ay_amp,porta_read,portb_read,porta_write,portb_write);
  //Inicializo el OPN
  FM2203[num].OPN:=opn_init(4);
  //Inicializo el state
  FM2203[num].OPN.type_:=TYPE_YM2203;
  FM2203[num].OPN.ST.clock:=clock;
  FM2203[num].OPN.ST.rate:=freq_base_audio;
  FM2203[num].tsample:=init_channel;
  fm2203[num].opn.ST.IRQ_Handler:=irq_func;
  YM2203_Reset(num);
  ym2203_init_timers(snd_clock,num_cpu,num);
end;

procedure YM2203_Update(num:byte);
var
  OPN:pfm_opn;
  lt:integer;
  cch:array[0..2] of pfm_chan;
begin
		OPN:=FM2203[num].OPN;
		cch[0]:=OPN.p_CH[0];
		cch[1]:=OPN.p_CH[1];
		cch[2]:=OPN.p_CH[2];
    // refresh PG and EG */
	  refresh_fc_eg_chan(OPN,cch[0]);
	  refresh_fc_eg_chan(OPN,cch[1]);
		if ((OPN.ST.mode and $c0)<>0) then begin
		  // 3SLOT MODE */
		  if (cch[2].SLOT[SLOT1].Incr=-1) then begin
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT1],OPN.SL3.fc[1],OPN.SL3.kcode[1] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT2],OPN.SL3.fc[2],OPN.SL3.kcode[2] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT3],OPN.SL3.fc[0],OPN.SL3.kcode[0] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT4],cch[2].fc,cch[2].kcode );
		  end;
    end else begin
      refresh_fc_eg_chan(OPN,cch[2]);
    end;
	  // YM2203 doesn't have LFO so we must keep these globals at 0 level */
	  LFO_AM:= 0;
	  LFO_PM:= 0;
	  // buffering */
    out_fm[0]:= 0;
    out_fm[1]:= 0;
    out_fm[2]:= 0;
		  // advance envelope generator */
    OPN.eg_timer:=OPN.eg_timer+OPN.eg_timer_add;
    while (OPN.eg_timer>=OPN.eg_timer_overflow) do begin
			  OPN.eg_timer:=OPN.eg_timer-OPN.eg_timer_overflow;
			  OPN.eg_cnt:=OPN.eg_cnt+1;
			  advance_eg_channel(OPN,cch[0]);
			  advance_eg_channel(OPN,cch[1]);
			  advance_eg_channel(OPN,cch[2]);
    end;
    // calculate FM */
    chan_calc(OPN,cch[0],0);
    chan_calc(OPN,cch[1],0);
    chan_calc(OPN,cch[2],0);
    // buffering */
    lt:=out_fm[0]+out_fm[1]+out_fm[2]+ay8910_update_internal(num)^;
    if lt>$7fff then lt:=$7fff
      else if lt<-$7fff then lt:=-$7fff;
    tsample[FM2203[num].tsample,sound_status.posicion_sonido]:=lt div 2;
    INTERNAL_TIMER_A(num,FM2203[num].OPN.ST,FM2203[num].OPN.p_ch[2]);
    INTERNAL_TIMER_B(num,FM2203[num].OPN.ST,1)
end;

procedure YM2203Write(num,port,data:byte);
var
  OPN:pfm_opn;
  addr:integer;
begin
		OPN:=FM2203[num].OPN;
		if ((port and 1)= 0 ) then begin
			// address port */
			OPN.ST.address:=data;
			// Write register to SSG emurator */
			if (data<16) then AY8910_Control(data,num);
			// prescaler select : 2d,2e,2f  */
		  if ((data>=$2d) and (data<=$2f)) then OPNPrescaler_w(OPN,data,1);
		end else begin
			// data port */
			addr:=OPN.ST.address;
      FM2203[num].REGS[addr]:=data;
			case (addr and $f0) of
			  $00:begin	// 0x00-0x0f : SSG section */
              // Write data to SSG emurator */
              AY8910_Write(data,num);
            end;
			  $20:begin	// 0x20-0x2f : Mode section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteMode(num,OPN,addr,data);
				    end;
			  else begin	// 0x30-0xff : OPN section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteReg(OPN,addr,data);
              end;
		  end;
    end;
end;

function YM2203Read(num,port:byte):byte;
var
  adr,ret:byte;
begin
ret:=0;
if ((port and 1)=0 ) then begin
  // status port */
  ret:=FM2203[num].OPN.ST.status;
end else begin
  //data port (ONLY SSG) */
  adr:=FM2203[num].OPN.ST.address;
  if (adr<16) then ret:=AY8910_Read(num);
end;
YM2203Read:=ret;
end;

function ym2203_read_status(num:byte):byte;
begin
  ym2203_read_status:=YM2203Read(num,0);
end;

function ym2203_read_reg(num:byte):byte;
begin
  ym2203_read_reg:=YM2203Read(num,1);
end;

procedure ym2203_control(num,data:byte);
begin
  YM2203Write(num,0,data);
end;

procedure ym2203_write_reg(num,data:byte);
begin
  YM2203Write(num,1,data);
end;

procedure ym2203_timer1_0;
begin
  TimerAOver(0,FM2203[0].OPN.ST);
  if (FM2203[0].OPN.ST.mode and $80)<>0 then begin
			// CSM mode auto key on */
			CSMKeyControll(FM2203[0].OPN.type_,FM2203[0].OPN.p_ch[2]);
  end;
end;

procedure ym2203_timer2_0;
begin
  TimerBOver(0,FM2203[0].OPN.ST);
end;

procedure ym2203_timer1_1;
begin
  TimerAOver(1,FM2203[1].OPN.ST);
  if (FM2203[1].OPN.ST.mode and $80)<>0 then begin
			// CSM mode auto key on */
			CSMKeyControll(FM2203[1].OPN.type_,FM2203[1].OPN.p_ch[2]);
  end;
end;

procedure ym2203_timer2_1;
begin
  TimerBOver(1,FM2203[1].OPN.ST);
end;

procedure YM2203_init_timer_a(num:byte;count:integer);
begin
  if count=0 then timer[FM2203[num].timer1].enabled:=false
    else begin
      timer[FM2203[num].timer1].enabled:=true;
      timer[FM2203[num].timer1].time_final:=count*FM2203[num].timer_adjust;
    end;
end;

procedure YM2203_init_timer_b(num:byte;count:integer);
begin
  if count=0 then timer[FM2203[num].timer2].enabled:=false
    else begin
      timer[FM2203[num].timer2].enabled:=true;
      timer[FM2203[num].timer2].time_final:=count*FM2203[num].timer_adjust;
    end;
end;

end.
